# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.cmake** provides ``"CMake"``, ``"Build"``, ``"Pack"`` and
``"Test"`` steps, as well as CMake-specific initialization context.
"""

from . import build, config, context, pack, test

__all__ = ["build", "config", "context", "pack", "test"]



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.flow** contains the inner workings of various *C++ flow*
components.
"""

from . import cli, configs, dependency, init, interact, layer, steps

__all__ = ["cli", "configs", "dependency", "init", "interact", "layer", "steps"]



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.api.env** provides interaction with project environment:

- :class:`FlowConfig` loads the project config in ``.flow/config.yml``.
- :class:`Runtime` provides means to run tools and print messages, while
  respecting ``--dry-run``, ``--silent`` and ``--verbose``.
- :class:`Config` represents single run configuration.
"""

import argparse
import importlib
import json
import os
import re
import shlex
import shutil
import subprocess
import sys
from dataclasses import dataclass
from typing import Dict, List, Optional, Union, cast

import yaml

from cxx_flow.base import uname
from cxx_flow.base.plugins import load_module_plugins

platform = uname.uname()[0]


_flow_config_default_compiler: Optional[Dict[str, str]] = None


def default_compiler():
    try:
        return os.environ["DEV_CXX"]
    except KeyError:
        pass

    try:
        return _flow_config_default_compiler[platform]
    except KeyError:
        print(
            f"-- KeyError: {platform} in {_flow_config_default_compiler}",
            file=sys.stderr,
        )
        return "?"
    except TypeError:
        print(f"-- TypeError: internal: flow config not ready yet", file=sys.stderr)
        return "?"


class Printer:
    @staticmethod
    def hide(arg: str, secrets: List[str]):
        for secret in secrets:
            arg = arg.replace(secret, "?" * max(15, len(secret)))
        return arg

    @staticmethod
    def print_arg(arg: str, secrets: List[str], raw: bool):
        color = ""
        arg = Printer.hide(arg, secrets)
        if arg[:1] == "-":
            color = "\033[2;37m"
        if not raw:
            arg = shlex.join([arg])
        if color == "" and arg[:1] in ["'", '"']:
            color = "\033[2;34m"
        if color == "":
            return arg
        return f"{color}{arg}\033[m"

    @staticmethod
    def print_cmd(*args: str, use_color: bool = True, secrets: List[str], raw: bool):
        cmd = args[0] if raw else shlex.join([args[0]])
        if not use_color:
            if raw:
                print(cmd, *(Printer.hide(arg) for arg in args[1:]), file=sys.stderr)
            else:
                print(
                    cmd,
                    shlex.join(Printer.hide(arg) for arg in args[1:]),
                    file=sys.stderr,
                )
            return

        args = " ".join([Printer.print_arg(arg, secrets, raw) for arg in args[1:]])
        print(f"\033[33m{cmd}\033[m {args}", file=sys.stderr)


@dataclass
class RunAlias:
    name: str
    doc: str
    steps: List[str]

    @staticmethod
    def from_json(name: str, alias: Union[dict, list]):
        if isinstance(alias, dict):
            doc: str = alias.get("doc", "")
            steps: List[str] = alias.get("steps", [])
        else:
            doc = ""
            steps = alias
        if not doc:
            doc = f'Shortcut for "run -s {",".join(steps)}"'

        return RunAlias(name, doc, steps)


class FlowConfig:
    _cfg: dict
    steps: list = []
    aliases: List[RunAlias] = []

    def __init__(self, root: str = "."):
        self.root = os.path.abspath(root)
        try:
            with open(
                os.path.join(self.root, ".flow", "config.yml"),
                encoding="UTF-8",
            ) as f:
                self._cfg = yaml.load(f, Loader=yaml.Loader)
        except FileNotFoundError:
            self._cfg = {}

        self._propagate_compilers()
        self._load_plugins()

    def _propagate_compilers(self):
        global _flow_config_default_compiler
        _flow_config_default_compiler = self.compiler_os_default

    def _load_plugins(self):
        std_plugins = importlib.import_module("cxx_flow.plugins")
        load_module_plugins(std_plugins)

        local_plugins = os.path.abspath(os.path.join(self.root, ".flow", "extensions"))
        if not os.path.isdir(local_plugins):
            plugins = importlib.import_module(local_plugins)
            load_module_plugins(plugins, can_fail=True)
            return

        sys.path.insert(0, local_plugins)

        for root, dirnames, _ in os.walk(local_plugins):
            for dirname in dirnames:
                init = os.path.join(root, dirname, "__init__.py")
                if not os.path.isfile(init):
                    continue
                plugins = importlib.import_module(dirname)
                load_module_plugins(plugins, can_fail=True)
            dirnames[:] = []

        sys.path.pop(0)

    @property
    def entry(self) -> Dict[str, dict]:
        return self._cfg.get("entry", {})

    @property
    def compiler(self) -> Dict[str, dict]:
        return self._cfg.get("compiler", {})

    @property
    def compiler_names(self) -> Dict[str, List[str]]:
        return self.compiler.get("names", {})

    @property
    def compiler_os_default(self) -> Dict[str, str]:
        return self.compiler.get("os-default", {})

    @property
    def lts_list(self) -> Dict[str, List[str]]:
        return self._cfg.get("lts", {})

    @property
    def postproc(self) -> dict:
        return self._cfg.get("postproc", {})

    @property
    def postproc_exclude(self) -> List[dict]:
        return self.postproc.get("exclude", [])

    @property
    def shortcuts(self) -> Dict[str, dict]:
        return self._cfg.get("shortcuts", {})


def _mkdir(dirname: str):
    os.makedirs(dirname, exist_ok=True)


def _ls(dirname: str, shallow=True):
    result = []
    for root, dirnames, filenames in os.walk(dirname):
        if shallow:
            dirnames[:] = []

        result.extend(
            os.path.relpath(os.path.join(root, filename), start=dirname)
            for filename in filenames
        )
    return result


def _cp(src: str, dst: str) -> int:
    try:
        dst = os.path.abspath(dst)
        if os.path.isdir(src):
            _mkdir(dst)
            shutil.copytree(src, dst, dirs_exist_ok=True, symlinks=True)
        else:
            if not os.path.isdir(dst):
                _mkdir(os.path.dirname(dst))
            shutil.copy(src, dst, follow_symlinks=False)
    except FileNotFoundError as err:
        print(err, file=sys.stderr)
        return 1


class Runtime(FlowConfig):
    dry_run: bool
    silent: bool
    verbose: bool
    official: bool
    no_coverage: bool
    use_color: bool
    only_host: bool
    platform: str
    secrets: List[str] = []

    def __init__(self, argsOrRuntime: Union[argparse.Namespace, "Runtime"]):
        super().__init__()

        if isinstance(argsOrRuntime, argparse.Namespace):
            args = argsOrRuntime
            self.dry_run = getattr(args, "dry_run", False)
            self.silent = getattr(args, "silent", False)
            self.verbose = getattr(args, "verbose", False)
            self.official = getattr(args, "official", False)
            self.use_color = True
            self.no_coverage = False
            self.platform = platform

            if "NO_COVERAGE" in os.environ:
                self.no_coverage = True

            if "RELEASE" in os.environ and "GITHUB_ACTIONS" in os.environ:
                self.official = not not json.loads(os.environ["RELEASE"])

            self.only_host = not (self.dry_run or self.official)
        else:
            rt = argsOrRuntime
            self.dry_run = rt.dry_run
            self.silent = rt.silent
            self.verbose = rt.verbose
            self.official = rt.official
            self.no_coverage = rt.no_coverage
            self.use_color = rt.use_color
            self.only_host = rt.only_host
            self.platform = rt.platform
            self.secrets = [*rt.secrets]

    def message(self, *args: str, **kwargs):
        if not self.verbose:
            return

        print("--", *args, **kwargs, file=sys.stderr)

    def print(self, *args: str, raw=False):
        if not self.silent:
            Printer.print_cmd(
                *args, use_color=self.use_color, secrets=self.secrets, raw=raw
            )

    def cmd(self, *args: str):
        self.print(*args)
        if self.dry_run:
            return 0

        result = subprocess.run(args)
        if result.returncode != 0:
            print(
                f"cxx-flow: error: {args[0]} ended in failure, exiting",
                file=sys.stderr,
            )
            return 1
        return 0

    def mkdirs(self, dirname: str):
        self.print("mkdir", "-p", dirname)

        if self.dry_run:
            return 0

        os.makedirs(dirname, exist_ok=True)
        return 0

    def cp(self, src: str, dst: str, regex: Optional[str] = None):
        args = ["cp"]
        if os.path.isdir(src):
            args.append("-r")
        self.print(*args, src, dst)

        if self.dry_run:
            return 0

        if regex is None:
            return _cp(src, dst)

        files = _ls(src)
        files = (name for name in files if re.match(regex, name))
        for name in files:
            result = _cp(os.path.join(src, name), os.path.join(dst, name))
            if result:
                return result
        return 0


@dataclass
class Config:
    items: dict
    keys: List[str]

    def get_path(self, key: str, val: any = None):
        path = key.split(".")
        ctx = self.items
        for step in path:
            if not isinstance(ctx, dict):
                return val
            child = ctx.get(step)
            if child is None:
                return val
            ctx = child
        return cast(any, ctx)

    @property
    def os(self) -> str:
        return self.items.get("os", "")

    @property
    def compiler(self) -> List[str]:
        return self.items.get("compiler", [])

    @property
    def build_type(self) -> str:
        return self.items.get("build_type", "")

    @property
    def build_name(self) -> str:
        return self.items.get("build_name", "")

    @property
    def build_dir(self) -> str:
        return os.path.join("build", self.preset)

    @property
    def preset(self) -> str:
        return self.items.get("preset", "")

    @property
    def build_generator(self) -> str:
        return self.items.get("build_generator", "")



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.flow.configs** provides :class:`Configs`, which uses
:class:`api.env.FlowConfig` to load the matrix definition and filter it out
using ``-D`` switches.
"""


import argparse
import os
from typing import Dict, List

from cxx_flow.api import env
from cxx_flow.base import matrix


def _compiler_inner(
    value: str, used_compilers: Dict[str, List[str]], config_names: Dict[str, List[str]]
):
    compiler, names = matrix.find_compiler(value, config_names)
    if compiler not in used_compilers:
        used_compilers[compiler] = []
    used_compilers[compiler].append(names)
    return compiler


def _compiler(used_compilers: Dict[str, List[str]], config_names: Dict[str, List[str]]):
    return lambda value: _compiler_inner(value, used_compilers, config_names)


def _boolean_inner(value: str, with_name: str):
    v = value.lower()
    return v in _TRUE or v == with_name


def _boolean(with_name: str):
    return lambda value: _boolean_inner(value, with_name)


_TRUE = {"true", "on", "yes", "1"}
_boolean_sanitizer = _boolean("with-sanitizer")


def _types(used_compilers: Dict[str, List[str]], config_names: Dict[str, List[str]]):
    return {
        "compiler": _compiler(used_compilers, config_names),
        "sanitizer": _boolean_sanitizer,
    }


def _config(config: List[str], only_host: bool, types: Dict[str, callable]):
    args = {}
    for arg in config:
        if arg[:1] == "-":
            continue
        _arg = arg.split("=", 1)
        if len(_arg) == 1:
            continue

        name, vals = _arg
        name = name.strip()
        conv = types.get(name, lambda value: value)
        values = {conv(val.strip()) for val in vals.split(",")}
        if name in args:
            values.update(args[name])
        args[name] = list(values)

    if only_host and "os" not in args:
        args["os"] = [env.platform]

    return args


def _expand_one(config: dict, github_os: str, os_in_name: str):
    os_ver = github_os.split("-")[1]
    build_name = f"{config['build_type']} with {config['compiler']} on {os_in_name}"
    if config["sanitizer"]:
        build_name += " (and sanitizer)"
    config["github_os"] = github_os
    config["build_name"] = build_name
    config["needs_gcc_ppa"] = os_ver != "latest" and config["os"] == "ubuntu"
    return config


def _expand_config(config: dict, spread_lts: bool, lts_list: Dict[str, List[str]]):
    if spread_lts:
        spread = lts_list.get(config["os"], [])
        if len(spread):
            return [
                _expand_one({key: config[key] for key in config}, lts, lts)
                for lts in spread
            ]
    return [_expand_one(config, f"{config['os']}-latest", config["os"])]


def _load_flow_data(rt: env.Runtime):
    root = ".flow"
    paths = [os.path.join(root, "matrix.yml")]
    if rt.official:
        paths.append(os.path.join(root, "official.yml"))
    configs, keys = matrix.load_matrix(*paths)

    if rt.no_coverage:
        for conf in configs:
            if "coverage" in conf:
                del conf["coverage"]

    return configs, keys


class Configs:
    usable: List[env.Config] = []

    def __init__(self, rt: env.Runtime, args: argparse.Namespace):
        super()
        configs, keys = _load_flow_data(rt)

        used_compilers: Dict[str, List[str]] = {}

        types = _types(used_compilers=used_compilers, config_names=rt.compiler_names)
        arg_configs = matrix.cartesian(
            _config(matrix.flatten(args.configs), rt.only_host, types)
        )

        # from commands/github
        spread_lts = hasattr(args, "matrix") and not not args.matrix

        turned = matrix.flatten(
            [
                _expand_config(config, spread_lts, rt.lts_list)
                for config in configs
                if len(arg_configs) == 0 or matrix.matches_any(config, arg_configs)
            ]
        )

        postproc_exclude = rt.postproc_exclude
        usable = [
            config
            for config in turned
            if len(postproc_exclude) == 0
            or not matrix.matches_any(config, postproc_exclude)
        ]

        self.usable = []
        for conf in usable:
            try:
                compilers = used_compilers[conf["compiler"]]
            except KeyError:
                fallback_compiler = matrix.find_compiler(
                    conf["compiler"], config_names=rt.compiler_names
                )
                compilers = [fallback_compiler[1]]
            for compiler in compilers:
                self.usable.append(
                    env.Config(
                        {
                            **conf,
                            "compiler": compiler,
                            "--orig-compiler": conf["compiler"],
                        },
                        keys,
                    )
                )



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.api.completers** defines :py:mod:`argcomplete` functions for
``-C``, ``--step`` and ``-D``.
"""

import os
from typing import Dict, List, Union, cast

import yaml

from cxx_flow import api


def cd_completer(prefix, **kwargs):
    target_dir = os.path.dirname(prefix)
    incomplete_part = os.path.basename(prefix)
    try:
        names = os.listdir(target_dir or ".")
    except Exception:
        return  # empty iterator

    for name in names:
        if not name.startswith(incomplete_part):
            continue
        full = os.path.join(target_dir, name)
        if not os.path.isdir(full):
            continue

        yield f"{full}{os.sep}"


def step_completer(prefix: str, parser, **kwargs):
    flow_cfg = cast(api.env.FlowConfig, parser.flow)

    comma_sep = prefix.split(",")
    start = ",".join(comma_sep[:-1])
    if len(comma_sep) > 1:
        start += ","

    current = comma_sep[-1].lower()
    for step in flow_cfg.steps:
        name = cast(str, step.name)
        if name.lower().startswith(current):
            candidate = current + name[len(current) :]
            yield start + candidate


def _str_arg(arg: Union[bool, str]):
    if isinstance(arg, bool):
        return "ON" if arg else "OFF"
    return str(arg)


def matrix_completer(prefix: str, parser, **kwargs):
    flow_cfg = cast(api.env.FlowConfig, parser.flow)

    matrix_yml = os.path.join(flow_cfg.root, ".flow", "matrix.yml")
    with open(matrix_yml, "r", encoding="UTF-8") as contents:
        data: Dict[str, List[any]] = yaml.load(contents, Loader=yaml.Loader).get(
            "matrix", {}
        )

    comma_sep = prefix.split(",")
    start = ",".join(comma_sep[:-1])
    if len(comma_sep) > 1:
        start += ","

    current = comma_sep[-1].split("=", 1)

    if len(current) == 1:
        completions: List[str] = []
        for key in data:
            if key.startswith(current[0]):
                completions.append(f"{start}{key}=")
        if len(completions) != 1:
            return completions

        current = completions[0].split("=", 1)

    try:
        args = data[current[0]]
    except KeyError:
        return

    completions: List[str] = []
    for arg in map(_str_arg, args):
        if arg.startswith(current[1]):
            completions.append(f"{start}{current[0]}={arg}")

    return completions



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.api.init** exposes APIs used by initialization extensions.
"""

from typing import List

from cxx_flow.api.env import Runtime


class InitStep:
    def platform_dependencies(self) -> List[str]:
        return []

    def postprocess(self, rt: Runtime, context: dict):
        pass


__steps: List[InitStep] = []


def register_init_step(step: InitStep):
    global __steps
    __steps.append(step)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.cmake.test** provides ``"Test"`` step.
"""

import os

from cxx_flow.api import env, step

from .__version__ import CMAKE_VERSION


@step.register
class CMakeTest:
    """Runs tests in the project using ``preset`` config."""

    name = "Test"
    runs_after = ["Build"]

    def platform_dependencies(self):
        return [f"cmake>={CMAKE_VERSION}", f"ctest>={CMAKE_VERSION}"]

    def is_active(self, config: env.Config, rt: env.Runtime) -> int:
        return os.path.isfile("CMakeLists.txt") and os.path.isfile("CMakePresets.json")

    def run(self, config: env.Config, rt: env.Runtime) -> int:
        return rt.cmd("ctest", "--preset", config.preset)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.commands.system** implements ``./flow system`` command.
"""

import platform
import sys
from typing import Annotated

from cxx_flow.api import arg, env
from cxx_flow.base import uname


@arg.command("system")
def main(
    format: Annotated[
        str,
        arg.Argument(
            help="Select, what format should be returned",
            choices=["props", "platform", "debug"],
        ),
    ]
):
    """Produce system information for CI pipelines"""

    node = platform.node()
    system, version, arch = uname.uname()

    if format == "props":
        print(f"-phost.name='{node}'")
        print(f"-pos='{system}'")
        if version is not None:
            print(f"-pos.version='{version}'")
        print(f"-parch='{arch}'")

    if format == "platform":
        version = "" if version is None else f"-{version}"
        print(f"{system}{version}-{arch}")

    if format == "debug":
        for name in [
            "uname",
            "machine",
            "node",
            "platform",
            "processor",
            "release",
            "system",
            "version",
        ]:
            print(name, platform.__dict__[name](), file=sys.stderr)

        print("-----", file=sys.stderr)
        print(f"node {node}", file=sys.stderr)
        print(f"os {system}", file=sys.stderr)
        if version is not None:
            print(f"version {version}", file=sys.stderr)
        print(f"machine {arch}", file=sys.stderr)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.cmake.parser** contains simple CMake parser.
"""

import os
import re
from typing import Iterator, List, NamedTuple, Optional

TOKENS = [
    ("COMMENT", r"#.*"),
    ("STR", r'"[^"]*"'),
    ("OPEN", r"\("),
    ("CLOSE", r"\)"),
    ("IDENT", r'[^ \t\r\n()#"]+'),
    ("WS", r"\s+"),
]


class Token(NamedTuple):
    type: str
    value: str
    offset: int


class Arg(NamedTuple):
    value: str
    offset: int


class Command(NamedTuple):
    name: str
    args: List[Arg]
    offset: int


class Project(NamedTuple):
    name: Arg
    version: Arg
    stability: Arg
    description: Arg

    @property
    def ver(self):
        return f"{self.version.value}{self.stability.value}"

    @property
    def pkg(self):
        return f"{self.name.value}-{self.ver}"

    @property
    def tag(self):
        return f"v{self.ver}"


def _token_stream(text: str) -> Iterator[Token]:
    tok_regex = "|".join("(?P<%s>%s)" % pair for pair in TOKENS)
    get_token = re.compile(tok_regex).match
    offset = 0

    mo = get_token(text)
    while mo is not None:
        token_type = mo.lastgroup
        if token_type not in ["COMMENT", "WS"]:
            value = mo.group(token_type)
            yield Token(type=token_type, value=value, offset=offset)
        offset = mo.end()
        mo = get_token(text, offset)


def _command(cmd: Token, stream: Iterator[Token]):
    result = Command(name=cmd.value, args=[], offset=cmd.offset)
    if next(stream)[0] != "OPEN":
        return result
    for tok in stream:
        if tok.type == "CLOSE":
            break
        elif tok.type == "OPEN":
            continue
        elif tok.type == "STR":
            result.args.append(Arg(value=tok.value[1:-1], offset=tok.offset + 1))
        elif tok.type == "IDENT":
            result.args.append(Arg(value=tok.value, offset=tok.offset))
        else:
            print(tok)

    return result


def _cmake(filename: str):
    commands: List[Command] = []

    with open(filename, "r", encoding="UTF-8") as f:
        stream = _token_stream(f.read())
    for tok in stream:
        if tok.type == "IDENT":
            commands.append(_command(tok, stream))

    return commands


def get_project(dirname: str) -> Project:
    commands = _cmake(os.path.join(dirname, "CMakeLists.txt"))

    project_name: Optional[Arg] = None
    version = Arg("0.1.0", -1)
    version_stability: Optional[Arg] = None
    description = Arg("", -1)

    for cmd in commands:
        if cmd.name == "project":
            project_name = cmd.args[0]
            args = cmd.args[1:]
            project_dict = {
                name.value: value for name, value in zip(args[::2], args[1::2])
            }
            version = project_dict.get("VERSION", version)
            description = project_dict.get("DESCRIPTION", "")
            if version_stability is not None:
                break
        elif cmd.name == "set":
            var_name = cmd.args[0]
            if var_name.value == "PROJECT_VERSION_STABILITY":
                version_stability = cmd.args[1]
                if project_name is not None:
                    break

    if project_name is None:
        project_name = Arg("", -1)
    if version_stability is None:
        version_stability = Arg("", -1)

    project_version = Project(
        name=project_name,
        version=version,
        stability=version_stability,
        description=description,
    )

    return project_version



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.conan._conan** adds support for both Conan v1 and v2.
"""

import shutil
from abc import ABC, abstractmethod
from typing import Callable, List, cast

from cxx_flow.api.env import Config, Runtime
from cxx_flow.base import cmd
from cxx_flow.flow.dependency import VER_REGEX


class conan(ABC):
    version: int

    def __init__(self, version: int = 1):
        self.version = version

    def settings(self, cfg: Config):
        result: List[str] = []
        for threshold, name in enumerate(
            ["conan_settings", "conan2_settings"],
        ):
            if self.version <= threshold:
                break
            result = cast(List[str], cfg.items.get(name, result))
        return result

    @abstractmethod
    def config(
        self,
        rt: Runtime,
        conan_output_dir: str,
        compiler_profile_name: str,
        build_type_profile_name: str,
    ) -> int: ...


class conan_1(conan):
    def __init__(self):
        super().__init__(1)

    def config(
        self,
        rt: Runtime,
        conan_output_dir: str,
        compiler_profile_name: str,
        build_type_profile_name: str,
    ) -> int:
        if rt.cmd(
            "conan",
            "profile",
            "new",
            "--detect",
            "--force",
            compiler_profile_name,
        ):
            return 1

        return rt.cmd(
            "conan",
            "install",
            "-if",
            conan_output_dir,
            "-of",
            conan_output_dir,
            "--build",
            "missing",
            "-pr:b",
            build_type_profile_name,
            "-pr:h",
            build_type_profile_name,
            ".",
        )


class conan_2(conan):
    def __init__(self):
        super().__init__(2)

    def config(
        self,
        rt: Runtime,
        conan_output_dir: str,
        compiler_profile_name: str,
        build_type_profile_name: str,
    ) -> int:
        if rt.cmd(
            "conan",
            "profile",
            "detect",
            "--force",
            "--name",
            compiler_profile_name,
        ):
            return 1

        return rt.cmd(
            "conan",
            "install",
            "-of",
            conan_output_dir,
            "--build",
            "missing",
            "-pr:b",
            build_type_profile_name,
            "-pr:h",
            build_type_profile_name,
            ".",
        )


def _conan_version():
    found = shutil.which("conan")
    if found is None:
        return 1

    proc = cmd.run(found, "--version", capture_output=True)
    if proc.returncode != 0:
        return 1
    m = VER_REGEX.search(proc.stdout)
    version = m.group(0) if m is not None else None
    if version is None:
        return 1
    chunks = [int(v.strip()) for v in version.split(".")]
    if len(chunks) < 1:
        return 1
    return chunks[0]


ctors: List[Callable[[], conan]] = [
    conan_1,
    conan_2,
]


def conan_api() -> conan:
    version = _conan_version()
    index = version - 1
    if index >= len(ctors):
        return ctors[-1]()
    return ctors[index]()



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.flow.init** supports the ``init`` command.
"""

from cxx_flow.api import ctx


def _fixup(settings: ctx.SettingsType, key: str, fixup: str, force=False):
    value = settings.get(key, "")
    if value != "" and not force:
        return

    value = ctx._build_fixup(settings, fixup)
    settings[key] = value


def _get_default(setting: ctx.Setting, settings: ctx.SettingsType):
    value = setting.calc_value(settings)
    if isinstance(value, list):
        return value[0]
    return value


def all_default():
    settings: ctx.SettingsType = {}

    for setting in ctx.defaults:
        value = _get_default(setting, settings)
        settings[setting.json_key] = value

    for setting in ctx.switches:
        value = _get_default(setting, settings)
        settings[setting.json_key] = value

    return settings


def fixup(settings: ctx.SettingsType):
    for setting in ctx.hidden:
        value = _get_default(setting, settings)
        if isinstance(value, bool) or value != "":
            settings[setting.json_key] = value

    for coll in [ctx.defaults, ctx.hidden]:
        for setting in coll:
            _fixup(settings, setting.json_key, setting.fix or "", setting.force_fix)
    del settings["EXT"]

    result = {}
    for key in settings:
        path = key.split(".")
        path_ctx = result
        for step in path[:-1]:
            if step not in path_ctx or not isinstance(path_ctx[step], dict):
                path_ctx[step] = {}
            path_ctx = path_ctx[step]
        path_ctx[path[-1]] = settings[key]
    return result


def get_internal(key: str, value: any = None):
    return ctx.internals.get(key, value)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.commands.run** implements ``./flow run`` command.
"""

import os
import shutil
import sys
from contextlib import contextmanager
from typing import Annotated, List, Optional, Set, cast

from cxx_flow import api
from cxx_flow.base import matrix
from cxx_flow.flow import dependency
from cxx_flow.flow.configs import Configs


@api.arg.command("run")
def main(
    steps: Annotated[
        Optional[List[str]],
        api.arg.Argument(
            help="Run only listed steps; if missing, run all the steps",
            names=["-s", "--steps"],
            nargs="*",
            meta="step",
            action="append",
            default=[],
            completer=api.completers.step_completer,
        ),
    ],
    configs: Configs,
    rt: api.env.Runtime,
):
    """Run automation steps for current project"""

    rt_steps = cast(List[api.step.Step], rt.steps)
    steps = matrix.flatten(step.split(",") for step in matrix.flatten(steps))
    if not steps:
        steps = [step.name for step in rt_steps]
    steps = list(map(lambda s: s.lower(), steps))

    step_names = set(steps)
    program = [step for step in rt_steps if step.name.lower() in step_names]

    errors = gather_dependencies_for_all_configs(configs, rt, program)
    if len(errors) > 0:
        if not rt.silent:
            for error in errors:
                print(f"cxx-flow: {error}", file=sys.stderr)
        return 1

    printed = refresh_directories(configs, rt, program)
    return run_steps(configs, rt, program, printed)


def gather_dependencies_for_all_configs(
    configs: Configs, rt: api.env.Runtime, steps: List[api.step.Step]
):
    deps: List[dependency.Dependency] = []
    for config in configs.usable:
        active_steps = [step for step in steps if step.is_active(config, rt)]
        deps.extend(dependency.gather(active_steps))
    return dependency.verify(deps)


def refresh_directories(
    configs: Configs, rt: api.env.Runtime, steps: List[api.step.Step]
):
    directories_to_refresh: Set[str] = set()
    for config in configs.usable:
        for step in steps:
            if step.is_active(config, rt):
                dirs = step.directories_to_remove(config)
                directories_to_refresh.update(dirs)

    printed = False
    for dirname in directories_to_refresh:
        if not rt.silent:
            printed = True
            print(f"[-] {dirname}", file=sys.stderr)
        if not rt.dry_run:
            shutil.rmtree(dirname, ignore_errors=True)

    return printed


COMPILER_ENV = ["CC", "CXX"]


@contextmanager
def compilers_env_setup(compiler: List[str], rt: api.env.Runtime):
    orig_env = {}
    if sys.platform != "win32":
        for var, value in zip(COMPILER_ENV, compiler):
            if var in os.environ:
                orig_env[var] = os.environ[var]
            os.environ[var] = value
            rt.message(f"set {var}={value}")
    try:
        yield
    finally:
        for var in COMPILER_ENV:
            if var in os.environ:
                del os.environ[var]
            if var not in orig_env:
                rt.message(f"unset {var}")
        for var, value in orig_env.items():
            os.environ[var] = value
            rt.message(f"set {var}={value}")


def run_steps(
    configs: Configs, rt: api.env.Runtime, program: List[api.step.Step], printed: bool
) -> int:
    config_count = len(configs.usable)
    for config_index in range(config_count):
        config = configs.usable[config_index]
        steps = [step for step in program if step.is_active(config, rt)]
        step_count = len(steps)
        if step_count == 0:
            continue

        if printed:
            print(file=sys.stderr)
        printed = True

        if config_count < 2:
            print(f"- {config.build_name}", file=sys.stderr)
        else:
            print(
                f"- {config_index + 1}/{config_count}: {config.build_name}",
                file=sys.stderr,
            )

        with compilers_env_setup(config.compiler, rt):
            for index in range(step_count):
                step = steps[index]
                print(f"-- step {index + 1}/{step_count}: {step.name}", file=sys.stderr)
                ret = step.run(config, rt)
                if ret:
                    return 1

    return 0



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.store** provides ``"Store"`` step.
"""

from cxx_flow.api import step

from .store_packages import StorePackages
from .store_tests import StoreTests


@step.register
class StoreBoth(step.SerialStep):
    """Stores all artifacts created for ``preset`` config value."""

    name = "Store"

    def __init__(self):
        super().__init__()
        self.children = [StoreTests(), StorePackages()]



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.cmake.context** provides CMake-specific initialization
context.
"""

import re
import uuid

import chevron

from cxx_flow import api, flow

from .__version__ import CMAKE_VERSION

config_json_mustache = """
{{#with_cmake}}
cmake:
  vars:
    {{NAME_PREFIX}}_COVERAGE: "?config:coverage"
    {{NAME_PREFIX}}_SANITIZE: "?config:sanitizer"
    {{NAME_PREFIX}}_CUTDOWN_OS: "?runtime:cutdown_os"

{{/with_cmake}}
"""


class CMakeInit(api.init.InitStep):
    def postprocess(self, rt: api.env.Runtime, context: dict):
        patch = chevron.render(config_json_mustache, context).rstrip()
        if not patch:
            return

        with open(".flow/config.yml", encoding="UTF-8") as config_file:
            patched = re.split(r"(\nshortcuts:\n)", config_file.read())

        if len(patched) != 3:
            return

        patched.insert(1, patch)
        content = "".join(patched)

        with open(".flow/config.yml", "w", encoding="UTF-8") as config_file:
            config_file.write(content)


def _list_cmake_types():
    return api.ctx.move_to_front(
        "console-application",
        sorted(key for key in flow.init.get_internal("cmake").keys() if key),
    )


api.init.register_init_step(CMakeInit())
api.ctx.register_init_setting(
    api.ctx.Setting("PROJECT.TYPE", "CMake project type", _list_cmake_types)
)
api.ctx.register_init_setting(
    api.ctx.Setting("cmake", fix="{PROJECT.TYPE$map:cmake}"),
    api.ctx.Setting("CMAKE_VERSION", value=CMAKE_VERSION),
    api.ctx.Setting("PROJECT.WIX.UPGRADE_GUID", value=lambda: str(uuid.uuid4())),
    is_hidden=True,
)
api.ctx.register_switch("with_cmake", "Use CMake", True)
api.ctx.register_internal(
    "cmake",
    {
        "": {"cmd": "add_executable", "type": ""},
        "console-application": {
            "cmd": "add_executable",
            "type": "",
            "console-application": True,
            "console": True,
            "application": True,
            "link_access": "PRIVATE",
        },
        "win32-application": {
            "cmd": "add_executable",
            "type": " WIN32",
            "win32-application": True,
            "win32": True,
            "application": True,
            "link_access": "PRIVATE",
        },
        "static-library": {
            "cmd": "add_library",
            "type": " STATIC",
            "static-library": True,
            "static": True,
            "library": True,
            "link_access": "PUBLIC",
        },
        "shared-library": {
            "cmd": "add_library",
            "type": " SHARED",
            "shared-library": True,
            "shared": True,
            "library": True,
            "link_access": "PUBLIC",
        },
        "plugin-library": {
            "cmd": "add_library",
            "type": " MODULE",
            "plugin-library": True,
            "plugin": True,
            "library": True,
            "link_access": "PUBLIC",
        },
    },
)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.api.arg** is used by various commands to declare CLI arguments.
"""

import argparse
import inspect
import typing
from dataclasses import dataclass, field


@dataclass
class Argument:
    help: str = ""
    pos: bool = False
    names: typing.List[str] = field(default_factory=list)
    nargs: typing.Union[str, int, None] = None
    opt: typing.Optional[bool] = None
    meta: typing.Optional[str] = None
    action: typing.Union[str, argparse.Action, None] = None
    default: typing.Optional[typing.Any] = None
    choices: typing.Optional[typing.List[str]] = None
    completer: typing.Optional[callable] = None


class FlagArgument(Argument):
    def __init__(self, help: str = "", names: typing.List[str] = []):
        super().__init__(
            help=help, names=names, opt=True, action="store_true", default=False
        )


@dataclass
class _Command:
    name: str
    entry: typing.Optional[callable]
    doc: typing.Optional[str]
    subs: typing.Dict[str, "_Command"]

    def add(self, names: typing.List[str], entry: callable, doc: typing.Optional[str]):
        name = names[0]
        rest = names[1:]
        if len(rest):
            try:
                child = self.subs[name]
            except KeyError:
                child = _Command(name, None, None, {})
                self.subs[name] = child

            child.add(rest, entry, doc)
            return

        try:
            child = self.subs[name]
            child.entry = entry
            child.doc = doc
        except KeyError:
            self.subs[name] = _Command(name, entry, doc, {})


_known_subcommand: typing.List[callable] = []
_known_commands = _Command("", None, None, {})


def flow_subcommand(entry: callable):
    global _known_subcommand
    _known_subcommand.append(entry)
    return entry


def get_subcommands():
    global _known_subcommand
    return _known_subcommand


_autodoc = {
    "cxx_flow.flow.configs.Configs": "Holds all current configurations.",
    "cxx_flow.api.env.Runtime": "Means to run tools and print messages, while respecting ``--dry-run``, ``--silent`` and ``--verbose``.",
}


def _type_name(t) -> str:
    if t == type(None):
        return "None"

    if type(t) == type:
        return t.__name__

    origin = typing.get_origin(t)

    if origin is typing.Union:
        seen: typing.Set[str] = set()
        names: typing.List[str] = []
        for arg in typing.get_args(t):
            for name in _union_args(arg):
                if name in seen:
                    continue
                names.append(name)
        return "|".join(names)

    if origin in [list, type, dict]:
        args = typing.get_args(t)
        arg_list = ", ".join(_type_name(arg) for arg in args)
        return f"{origin.__name__}[{arg_list}]"

    args = typing.get_args(t)
    arg_list = ", ".join(_type_name(arg) for arg in args)
    if arg_list:
        return f"{origin.__name__}[{arg_list}]"

    return typing.cast(str, origin.__name__)

def _union_args(t) -> typing.Generator[str, None, None]:
    origin = typing.get_origin(t)

    if origin is typing.Union:
        args = typing.get_args(t)
        for arg in args:
            for candidate in _union_args(arg):
                yield candidate
        return

    yield _type_name(t)
    

def command(*name: str):
    def wrap(entry: callable):
        global _known_commands
        _known_commands.add(list(name), entry, entry.__doc__)

        doc = inspect.getdoc(entry) or ""
        if doc:
            doc += "\n\n"

        signature = inspect.signature(entry)

        for param_name, param in signature.parameters.items():
            annotation = param.annotation

            if type(annotation) == type:
                full_name = f"{annotation.__module__}.{annotation.__name__}"
                help = _autodoc.get(full_name, "")
                doc += f":param {annotation.__name__} {param_name}: {help}\n"
                continue

            if typing.get_origin(annotation) is not typing.Annotated:
                continue

            anno_args = typing.get_args(annotation)
            if len(anno_args) == 0:
                continue
            
            origin = anno_args[0]

            help = ""
            for arg in anno_args[1:]:
                if isinstance(arg, Argument):
                    help = arg.help
                    if help:
                        break

            doc += f":param {_type_name(origin)} {param_name}: {help}\n"

        entry.__doc__ = doc

        return entry

    return wrap


def get_commands():
    return _known_commands



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.store** provides ``"StorePackages"`` step.
"""

import os
import shutil
from typing import List, cast

from cxx_flow.api import env, step
from cxx_flow.base.uname import uname

from ..cmake.parser import get_project

_system, _version, _arch = uname()
_version = "" if _version is None else f"-{_version}"
_project_pkg = None


def _package_name(config: env.Config, pkg: str, group: str):
    debug = "-dbg" if config.build_type.lower() == "debug" else ""
    suffix = group and f"-{group}" or ""

    return f"{pkg}-{_system}{_version}-{_arch}{debug}{suffix}"


@step.register
class StorePackages:
    """Stores archives and installers build for ``preset`` config value."""

    name = "StorePackages"
    runs_after = ["Pack"]

    def run(self, config: env.Config, rt: env.Runtime) -> int:
        if not rt.dry_run:
            os.makedirs("build/artifacts", exist_ok=True)

        packages_dir = f"build/{config.preset}/packages"

        global _project_pkg
        if _project_pkg is None:
            _project_pkg = get_project("").pkg

        main_group = cast(List[str], rt._cfg.get("package", {}).get("main-group"))
        if main_group is not None and not rt.dry_run:
            src = _package_name(config, _project_pkg, main_group)
            dst = _package_name(config, _project_pkg, "")
            rt.print("mv", *(f"{package}.*" for package in (src, dst)), raw=True)
            for _, dirnames, filenames in os.walk(packages_dir):
                dirnames[:] = []
                extensions = [
                    filename[len(src) :]
                    for filename in filenames
                    if len(filename) > len(src)
                    and filename[: len(src)] == src
                    and filename[len(src)] == "."
                ]
            for extension in extensions:
                shutil.move(
                    f"{packages_dir}/{src}{extension}",
                    f"{packages_dir}/{dst}{extension}",
                )

        GITHUB_OUTPUT = os.environ.get("GITHUB_OUTPUT")
        if GITHUB_OUTPUT is not None:
            with open(GITHUB_OUTPUT, "a", encoding="UTF-8") as github_output:
                generators = ",".join(config.items.get("cpack_generator", []))
                print(f"CPACK_GENERATORS={generators}", file=github_output)

        return rt.cp(
            packages_dir,
            "build/artifacts/packages",
            f"^{_project_pkg}-.*$",
        )



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.cmake.build** provides ``"Build"`` step.
"""

import os

from cxx_flow.api import env, step

from .__version__ import CMAKE_VERSION


@step.register()
class CMakeBuild:
    """Builds the project using ``preset`` config."""

    name = "Build"
    runs_after = ["Conan", "CMake"]

    def platform_dependencies(self):
        return [f"cmake>={CMAKE_VERSION}"]

    def is_active(self, config: env.Config, rt: env.Runtime) -> int:
        return os.path.isfile("CMakeLists.txt") and os.path.isfile("CMakePresets.json")

    def run(self, config: env.Config, rt: env.Runtime) -> int:
        return rt.cmd("cmake", "--build", "--preset", config.preset, "--parallel")



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.commands.list** implements ``./flow list`` command.
"""

import os
from typing import Annotated, Dict, List, Optional, Set, cast

from cxx_flow import flow
from cxx_flow.api import arg, env, step
from cxx_flow.base import matrix


@arg.command("list")
def main(
    builtin: Annotated[bool, arg.FlagArgument(help="Show all builtin commands")],
    alias: Annotated[bool, arg.FlagArgument(help="Show all alias commands")],
    steps: Annotated[bool, arg.FlagArgument(help="Show all run steps")],
    configs: Annotated[bool, arg.FlagArgument(help="Show all known matrix keys")],
    all: Annotated[
        bool, arg.FlagArgument(help="Show builtins, aliases, steps and configs")
    ],
    pipe: Annotated[bool, arg.FlagArgument(help="Do not show additional information")],
    rt: env.Runtime,
):
    """List all the commands and/or steps for cxx-flow"""

    printed_something = False
    bold = "\033[96m" if rt.use_color else ""
    reset = "\033[0m" if rt.use_color else ""

    if all:
        builtin = True
        alias = True
        steps = True
        configs = True

    if builtin:
        builtin_entries = list(
            sorted((entry.name, entry.doc) for entry in flow.cli.cmds.command_list)
        )
        if not pipe and len(builtin_entries) > 0:
            print("Builtin commands")
            print("----------------")

        for entry_name, entry_doc in builtin_entries:
            if pipe:
                print(entry_name)
                continue

            name = f"{bold}{entry_name}{reset}"
            if entry_doc:
                print(f"- {name}: {entry_doc}")
            else:
                print(f"- {name}")

        printed_something = True

    if alias:
        aliases = rt.aliases

        if not pipe and len(aliases) > 0:
            if printed_something:
                print()

            print("Known aliases")
            print("-------------")

        for alias in aliases:
            if pipe:
                print(alias.name)
                continue

            name = f"{bold}{alias.name}{reset}"
            print(f"- {name}: {', '.join(alias.steps)}")

        printed_something = True

    if steps:
        rt_steps = cast(List[step.Step], rt.steps)

        if not pipe and len(rt_steps) > 0:
            if printed_something:
                print()

            print("Run steps")
            print("---------")

        some_unused = False
        aliased_steps = set(matrix.flatten([alias.steps for alias in rt.aliases]))

        for rt_step in rt_steps:
            if pipe:
                print(rt_step.name)
                continue

            step_used = rt_step.name in aliased_steps
            if not step_used:
                some_unused = True

            name = f"{bold}{rt_step.name}{reset}"
            if step_used:
                print(f"- {name}")
            else:
                print(f"- {name}*")

        if some_unused:
            print(
                f"*step can only be run by explicitly calling through {bold}run{reset}."
            )

        printed_something = True

    if configs:
        m, keys = _load_flow_data(rt)
        if pipe:
            for key in keys:
                print(key)
        else:
            if len(keys) > 0:
                if printed_something:
                    print()

                print("Matrix keys")
                print("-----------")

            values: Dict[str, Set[str]] = {}

            for config in m:
                for key in keys:
                    value = config[key]
                    if isinstance(value, bool):
                        value = "ON" if value else "OFF"
                    else:
                        value = str(value)
                    try:
                        values[key].add(value)
                    except KeyError:
                        values[key] = {value}

            empty = set()
            for key in keys:
                value = ", ".join(values.get(key, empty))
                name = f"{bold}{key}{reset}"
                if value:
                    print(f"- {name}: {value}")
                else:
                    print(f"- {name}")

        printed_something = True

    if not printed_something and not pipe:
        print(f"Use {bold}--help{reset} to see, which listings are available")


def _load_flow_data(rt: env.Runtime):
    paths = [os.path.join(".flow", "matrix.yml")]
    m, keys = matrix.load_matrix(*paths)

    if rt.no_coverage:
        for conf in m:
            if "coverage" in conf:
                del conf["coverage"]

    return m, keys



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.commands.init** implements ``cxx-flow init`` command.
"""

import json
import os
import sys
from typing import Annotated, Optional

from cxx_flow import flow
from cxx_flow.api import arg, ctx, env, init


@arg.command("init")
def main(
    path: Annotated[
        Optional[str],
        arg.Argument(
            help="Location of initialized project. "
            "The directory will be created, if it does not exist yet. "
            "Defaults to current directory.",
            pos=True,
            default=".",
        ),
    ],
    non_interactive: Annotated[
        bool,
        arg.FlagArgument(help="Selects all the default answers", names=["-y", "--yes"]),
    ],
    save_context: Annotated[
        bool,
        arg.FlagArgument(help="Save the mustache context as JSON", names=["--ctx"]),
    ],
    rt: env.Runtime,
):
    """Initialize new project"""

    if path is not None:
        os.makedirs(path, exist_ok=True)
        os.chdir(path)

    errors = flow.dependency.verify(flow.dependency.gather(init.__steps))
    if len(errors) > 0:
        if not rt.silent:
            for error in errors:
                print(f"cxx-flow: {error}", file=sys.stderr)
        return 1

    context = flow.init.fixup(
        flow.init.all_default() if non_interactive else flow.interact.prompt()
    )
    if not non_interactive and not rt.silent:
        print()

    if save_context:
        with open(".context.json", "w", encoding="UTF-8") as jsonf:
            json.dump(context, jsonf, ensure_ascii=False, indent=4)

    flow.layer.copy_license(rt, context)
    if not rt.silent:
        print()

    layers = flow.layer.gather_package_layers(ctx.package_root, context)
    for fs_layer in layers:
        fs_layer.run(rt, context)

    if save_context:
        with open(".gitignore", "ab") as ignoref:
            ignoref.write("\n/.context.json\n".encode("UTF-8"))

    for step in init.__steps:
        step.postprocess(rt, context)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)
"""
The **cxx_flow.plugins.github** provides GitHub-related switches for new
projects.
"""

from cxx_flow.api import ctx

ctx.register_switch("with_github_actions", "Use Github Actions", True)
ctx.register_switch(
    "with_github_social", "Use Github ISSUE_TEMPLATE, CONTRIBUTING.md, etc.", True
)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.base** provides basic initialization setup for new
projects.
"""

from cxx_flow import __version__, api


class GitInit(api.init.InitStep):
    def platform_dependencies(self):
        return ["git"]

    def postprocess(self, rt: api.env.Runtime, context: dict):
        def git(*args):
            rt.cmd("git", *args)

        git("init")
        git("add", ".")
        git("commit", "-m", "Initial commit")


api.init.register_init_step(GitInit())
api.ctx.register_init_setting(
    api.ctx.Setting("__flow_version__", value=__version__),
    api.ctx.Setting("${", value="${"),
    is_hidden=True,
)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.base.cmd** defines environment for `cxx_flow.api.env.Runtime`.
"""

import shutil
import subprocess
from typing import Optional


def which(cmd: str) -> Optional[str]:
    """
    Return executable path from ``PATH`` environment variable, if it is present.

    :param cmd: name of the tool to expand
    :returns: the path to an executable which would be run if the given ``cmd``
       was called. If no ``cmd`` would be called, return ``None``.
    """
    return shutil.which(cmd)


def is_tool(name: str) -> bool:
    """
    :param name: name of the tool to check
    :returns: ``True``, if :func:`which` would return a string and `False`
       otherwise
    """
    return which(name) is not None


def run(app: str, *args: str, capture_output=False):
    """
    Run command with arguments and return a :py:class:`subprocess.CompletedProcess`
    instance. Internally, this function runs :py:func:`subprocess.run` with
    encoding set to "UTF-8" and check turned off.

    :param app: name of the command to run
    :param args: additional arguments of the command
    :param capture_output: if True, both stdout and stderr are captured
    :returns: The ``CompletedProcess`` instance if the command was successfully
       ran, or ``None``.
    """
    cmd = which(app)
    if cmd is None:
        return None
    return subprocess.run(
        [cmd, *args], check=False, encoding="UTF-8", capture_output=capture_output
    )



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.flow.layer** supports :ref:`template` layers.
"""

import json
import os
import shutil
from dataclasses import dataclass
from typing import List, Optional, cast

import chevron

from cxx_flow.api import ctx, env


@dataclass
class FileInfo:
    src: str
    dst: str
    is_mustache: bool
    when: Optional[str] = None

    @classmethod
    def from_json(cls, src: str, filelist: dict, context: ctx.SettingsType):
        basename, ext = os.path.splitext(src)
        is_mustache = ext == ".mustache"
        key = src.replace(os.sep, "/")
        json_file = cast(dict, filelist.get(key, {}))
        path = cast(Optional[str], json_file.get("path"))
        when = cast(Optional[str], json_file.get("when"))
        dst = (
            chevron.render(path, context).replace("/", os.sep)
            if path is not None
            else basename if is_mustache else src
        )
        return cls(src=src, dst=dst, is_mustache=is_mustache, when=when)

    def template(self):
        open_mstch = "{{"
        close_mstch = "}}"
        if self.when:
            return f"{open_mstch}#{self.when}{close_mstch}\n{self.dst}\n{open_mstch}/{self.when}{close_mstch}\n"
        return f"{self.dst}\n"

    def run(self, root: str, rt: env.Runtime, context: ctx.SettingsType):
        if not rt.silent:
            if rt.use_color:
                print(f"\033[2;30m+\033[m {self.dst}")
            else:
                print(f"+ {self.dst}")
        if rt.dry_run:
            return

        src = os.path.join(root, self.src)
        dst = os.path.abspath(self.dst)
        dirname = os.path.dirname(dst)
        os.makedirs(dirname, exist_ok=True)

        if self.is_mustache:
            with open(src, "rb") as inf:
                content = inf.read().decode("UTF-8")
            content = chevron.render(content, context)
            with open(dst, "wb") as outf:
                outf.write(content.encode("UTF-8"))
            shutil.copymode(src, dst, follow_symlinks=False)
            shutil.copystat(src, dst, follow_symlinks=False)
        else:
            shutil.copy2(src, dst, follow_symlinks=False)


@dataclass
class LayerInfo:
    root: str
    files: List[FileInfo]
    when: Optional[str] = None

    @classmethod
    def from_fs(cls, layer_dir: str, context: ctx.SettingsType):
        with open(f"{layer_dir}.json", encoding="UTF-8") as f:
            layer_info: dict = json.load(f)
        when = cast(Optional[bool], layer_info.get("when"))
        filelist = cast(dict, layer_info.get("filelist", {}))

        sources: List[str] = []
        prefix = os.path.join(layer_dir, "")
        for curr_dir, dirs, files in os.walk(layer_dir):
            dirs[:] = [dirname for dirname in dirs if dirname not in ["__pycache__"]]
            files[:] = [
                filename
                for filename in files
                if os.path.splitext(filename) not in [".pyc", ".pyo", ".pyd"]
            ]

            for filename in files:
                src = os.path.join(curr_dir, filename)
                if src.startswith(prefix):
                    src = src[len(prefix) :]
                sources.append(src)

        files = [FileInfo.from_json(src, filelist, context) for src in sources]
        files.sort(key=lambda info: info.dst)

        result = cls(root=layer_dir, files=files, when=when)
        allowed_files = set(
            filter(
                lambda path: path != "",
                chevron.render(result.template(), context).split("\n"),
            )
        )
        result.files = list(filter(lambda file: file.dst in allowed_files, files))

        return result

    @property
    def name(self):
        return os.path.basename(self.root)

    @property
    def pkg(self):
        return os.path.basename(
            os.path.dirname(
                os.path.dirname(os.path.dirname(os.path.dirname(self.root)))
            )
        )

    def template(self):
        open_mstch = "{{"
        close_mstch = "}}"
        result = ""
        if self.when:
            result += f"{open_mstch}#{self.when}{close_mstch}\n"
        result += "".join(file.template() for file in self.files)
        if self.when:
            result += f"{open_mstch}/{self.when}{close_mstch}\n"
        return result

    def run(self, rt: env.Runtime, context: ctx.SettingsType):
        if not rt.silent:
            if rt.use_color:
                print(f"\033[2;30m[{self.pkg}:{self.name}]\033[m")
            else:
                print(f"[{self.pkg}:{self.name}]")
        for file in self.files:
            file.run(self.root, rt, context)
        if not rt.silent:
            print()


def copy_license(rt: env.Runtime, context: ctx.SettingsType):
    license = context.get("COPY", {}).get("LICENSE")
    if not license:
        return

    licenses_dir = os.path.abspath(
        os.path.join(ctx.package_root, ctx.template_dir, "licenses")
    )
    license_file = f"{license}.mustache"
    info = FileInfo(license_file, "LICENSE", is_mustache=True)
    info.run(licenses_dir, rt, context)


def gather_package_layers(package_root: str, context: ctx.SettingsType):
    layers_dir = os.path.abspath(os.path.join(package_root, ctx.template_dir, "layers"))
    for root, dirs, files in os.walk(layers_dir):
        layer_dirs = [
            os.path.abspath(os.path.join(root, dirname))
            for dirname in dirs
            if f"{dirname}.json" in files
        ]
        dirs[:] = []

    layers = (LayerInfo.from_fs(layer_dir, context) for layer_dir in layer_dirs)
    return list(filter(lambda layer: len(layer.files) > 0, layers))



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.commands.bootstrap** implements ``./flow bootstrap`` command.
"""

import os

from cxx_flow.api import arg, env


@arg.command("bootstrap")
def main():
    """Finish bootstrapping on behalf of flow.py"""

    GITHUB_ENV = os.environ.get("GITHUB_ENV")
    if GITHUB_ENV is not None:
        with open(GITHUB_ENV, "a", encoding="UTF-8") as github_env:
            PATH = os.environ["PATH"]
            print(f"PATH={PATH}", file=github_env)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.flow.cli.cmds** provides command-line builders and runners,
supporting the functions defined in :mod:`cxx_flow.commands`.
"""

import argparse
import inspect
import typing
from dataclasses import dataclass, field
from typing import Annotated, Any, Dict, List, Optional, Tuple, Union, cast

from cxx_flow import __version__
from cxx_flow.api import arg, completers, env
from cxx_flow.flow.configs import Configs


def build_argparser(flow_cfg: env.FlowConfig):
    parser = argparse.ArgumentParser(
        prog="cxx-flow",
        description="C++ project maintenance, automated",
        add_help=False,
    )
    parser.add_argument(
        "-h", "--help", action="help", help="Show this help message and exit"
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        default=argparse.SUPPRESS,
        version=f"%(prog)s version {__version__}",
        help="Show cxx-flow's version and exit",
    )
    parser.add_argument(
        "-C",
        metavar="dir",
        nargs="?",
        help="Run as if cxx-flow was started in <dir> instead of the current "
        "working directory. This directory must exist.",
    ).completer = completers.cd_completer

    BuiltinEntry.visit_all(parser, flow_cfg)
    return parser


@dataclass
class SpecialArg:
    name: str
    ctor: callable

    def create(self, rt: env.Runtime, args: argparse.Namespace):
        if self.ctor == env.Runtime:
            return rt
        return self.ctor(rt, args)


@dataclass
class EntryArg:
    name: str
    opt: bool
    help: str
    pos: bool
    names: List[str]
    nargs: Union[str, int, None]
    meta: Optional[str]
    action: Union[str, argparse.Action, None]
    default: Optional[Any]
    choices: Optional[List[str]] = None
    completer: Optional[callable] = None

    def visit(self, parser: argparse.ArgumentParser):
        kwargs = {}
        if self.help is not None:
            kwargs["help"] = self.help
        if self.nargs is not None:
            kwargs["nargs"] = self.nargs
        if self.meta is not None:
            kwargs["metavar"] = self.meta
        if self.default is not None:
            kwargs["default"] = self.default
        if self.action is not None:
            kwargs["action"] = self.action
        if self.choices is not None:
            kwargs["choices"] = self.choices

        names = (
            [self.name]
            if self.pos
            else self.names if len(self.names) > 0 else [f"--{self.name}"]
        )

        if self.pos:
            kwargs["nargs"] = "?" if self.opt else 1
        else:
            kwargs["dest"] = self.name
            kwargs["required"] = not self.opt

        action = parser.add_argument(*names, **kwargs)
        if self.completer:
            action.completer = self.completer


@dataclass
class BuiltinEntry:
    name: str
    doc: str
    entry: callable
    args: List[EntryArg]
    additional: List[SpecialArg]
    children: List["BuiltinEntry"] = field(default_factory=list)

    @staticmethod
    def run_entry(args: argparse.Namespace, cfg: env.FlowConfig):
        builtin_entries = {entry.name for entry in command_list}
        aliases = cfg.aliases

        rt = env.Runtime(args)
        rt.steps = cfg.steps
        rt.aliases = cfg.aliases
        rt._cfg = cfg._cfg

        if args.command in builtin_entries:
            command = next(
                filter(lambda command: command.name == args.command, command_list)
            )
            return cast(BuiltinEntry, command).run(args, rt)
        elif args.command in {alias.name for alias in aliases}:
            command = next(filter(lambda command: command.name == "run", command_list))
            alias = next(filter(lambda alias: alias.name == args.command, aliases))
            args.steps.append(alias.steps)
            return cast(BuiltinEntry, command).run(args, rt)

        print("known commands:")
        for command in command_list:
            print(f"   {command.name}: {command.doc}")
        for alias in aliases:
            print(f"   {alias.name}: {alias.doc}")
        return 1

    def run(self, args: argparse.Namespace, rt: env.Runtime, level=0):
        if level == 0 and rt.only_host:
            rt.only_host = self.name == "run"

        subcommand_name = None

        if len(self.children):
            subcommand_attribute = f"command_{level}"
            if hasattr(args, subcommand_attribute):
                subcommand_name = getattr(args, subcommand_attribute)

        if subcommand_name is not None:
            subcommand = cast(
                BuiltinEntry,
                next(
                    filter(
                        lambda command: command.name == subcommand_name,
                        self.children,
                    )
                ),
            )
            return subcommand.run(args, rt, level=level + 1)

        kwargs = {}
        for arg in self.args:
            kwargs[arg.name] = getattr(args, arg.name, None)

        for additional in self.additional:
            arg = additional.create(rt, args)
            kwargs[additional.name] = arg

        result = self.entry(**kwargs)
        return 0 if result is None else result

    @staticmethod
    def visit_all(
        parser: argparse.ArgumentParser, cfg: env.FlowConfig
    ) -> Dict[str, List[str]]:
        global command_list
        command_list = BuiltinEntry.build_menu(arg.get_commands().subs)
        shortcut_configs = BuiltinEntry.build_shortcuts(cfg)

        parser.flow = cfg
        parser.shortcuts = shortcut_configs

        subparsers = parser.add_subparsers(
            dest="command", metavar="{command}", help="Known command name, see below"
        )

        subparsers.parent = parser

        run: BuiltinEntry = None
        for entry in command_list:
            entry.visit(subparsers)
            if entry.name == "run":
                run = entry

        if run is not None and len(cfg.aliases) > 0:
            builtin_entries = {entry.name for entry in command_list}
            cfg.aliases = [
                alias for alias in cfg.aliases if alias.name not in builtin_entries
            ]
            for alias in cfg.aliases:
                run.visit(subparsers, alias=alias.name, doc=alias.doc)
        else:
            cfg.aliases = []

    def visit(
        self,
        subparsers,
        alias: Optional[str] = None,
        doc: Optional[str] = None,
        level=0,
    ):
        if not doc:
            doc = self.doc
        if not alias:
            alias = self.name

        parser: argparse.ArgumentParser = subparsers.add_parser(
            alias, help=doc.split("\n\n")[0], description=doc, add_help=False
        )

        parent = getattr(subparsers, "parent")
        parser.flow = getattr(parent, "flow")
        parser.shortcuts = getattr(parent, "shortcuts")

        assert parent.flow is not None
        assert parent.shortcuts is not None

        parser.add_argument(
            "-h",
            "--help",
            action="help",
            default=argparse.SUPPRESS,
            help="Show this help message and exit",
        )

        parser.add_argument(
            "--dry-run",
            action="store_true",
            required=False,
            help="Print steps and commands, do nothing",
        )

        verbosity = parser.add_mutually_exclusive_group()
        verbosity.add_argument(
            "--silent",
            action="store_true",
            required=False,
            help="Remove most of the output",
        )
        verbosity.add_argument(
            "--verbose",
            action="store_true",
            required=False,
            help="Add even more output",
        )

        has_config = False
        for additional in self.additional:
            if additional.ctor == Configs:
                has_config = True
                break

        if has_config:
            parser.add_argument(
                "-D",
                dest="configs",
                metavar="key=value",
                nargs="*",
                action="append",
                default=[],
                help="Run only builds on matching configs. The key is one of "
                'the keys into "matrix" object in .flow/matrix.yml definition '
                "and the value is one of the possible values for that key. In "
                "case of boolean flags, such as sanitizer, the true value is "
                'one of "true", "on", "yes", "1" and "with-<key>", '
                'i.e. "with-sanitizer" for sanitizer.'
                " "
                "If given key is never used, all values from .flow/matrix.yaml "
                "for that key are used. Otherwise, only values from command "
                "line are used.",
            ).completer = completers.matrix_completer

            parser.add_argument(
                "--official",
                action="store_true",
                required=False,
                help="Cut matrix to release builds only",
            )

            if len(parser.shortcuts):
                group = parser.add_mutually_exclusive_group()

                for shortcut_name in sorted(parser.shortcuts.keys()):
                    config = parser.shortcuts[shortcut_name]
                    group.add_argument(
                        f"--{shortcut_name}",
                        required=False,
                        action="store_true",
                        help=f'Shortcut for "-D {" ".join(config)}"',
                    )

        for arg in self.args:
            arg.visit(parser)

        if len(self.children):
            subparsers = parser.add_subparsers(
                dest=f"command_{level}",
                metavar="{command}",
                help="Known command name, see below",
            )
            subparsers.parent = parser

            for entry in self.children:
                entry.visit(subparsers, level=level + 1)

    @staticmethod
    def build_shortcuts(cfg: env.FlowConfig) -> Dict[str, List[str]]:
        shortcut_configs: Dict[str, List[str]] = {}
        args: List[Tuple[str, List[str], bool, bool]] = []

        shortcuts = cfg.shortcuts
        for shortcut_name in sorted(shortcuts.keys()):
            has_os = False
            has_compiler = False
            shortcut = shortcuts[shortcut_name]
            config: List[str] = []
            for key in sorted(shortcut.keys()):
                has_os = has_os or key == "os"
                has_compiler = has_compiler or key == "os"
                value = shortcut[key]
                if isinstance(value, list):
                    for v in value:
                        config.append(f"{key}={_shortcut_value(v)}")
                else:
                    config.append(f"{key}={_shortcut_value(value)}")
            if len(config) > 0:
                args.append((shortcut_name, config, has_os, has_compiler))

        if len(args):
            os_prefix = f"os={env.platform}"
            compiler_prefix = f"compiler={env.default_compiler()}"

            for shortcut_name, config, has_os, has_compiler in args:
                if not has_compiler:
                    config.insert(0, compiler_prefix)
                if not has_os:
                    config.insert(0, os_prefix)
                shortcut_configs[shortcut_name] = config

        return shortcut_configs

    @staticmethod
    def build_menu(commands: Dict[str, arg._Command]):
        result: List[BuiltinEntry] = []
        for cmd in commands.values():
            name = cmd.name
            doc = cmd.doc or ""
            entry = cmd.entry or (lambda: 0)
            children = BuiltinEntry.build_menu(cmd.subs)

            args = _extract_args(entry)
            special_args = [entry for entry in args if isinstance(entry, SpecialArg)]
            entry_args = [entry for entry in args if isinstance(entry, EntryArg)]

            result.append(
                BuiltinEntry(
                    name=name,
                    doc=doc,
                    entry=entry,
                    args=entry_args,
                    additional=special_args,
                    children=children,
                )
            )

        return result


def _shortcut_value(value) -> str:
    if isinstance(value, bool):
        return "ON" if value else "OFF"
    return str(value)


def _extract_arg(name: str, argument: Any):
    for ctor in [Configs, env.Runtime]:
        if argument is ctor:
            return SpecialArg(name, ctor)

    anno_type = typing.get_origin(argument)
    if anno_type is not Annotated:
        return None

    arg_dict = getattr(argument, "__dict__", {})
    origin = arg_dict.get("__origin__", None)
    (metadata,) = arg_dict.get("__metadata__", None)

    if origin is None or not isinstance(metadata, arg.Argument):
        return None

    optional = metadata.opt
    if optional is None:
        optional = typing.get_origin(origin) is Union and type(None) in typing.get_args(
            origin
        )

    kwargs = metadata.__dict__.copy()
    del kwargs["opt"]
    return EntryArg(name, optional, **kwargs)


def _extract_args(entry: callable) -> List[Union[EntryArg, SpecialArg]]:
    entry_args = inspect.get_annotations(entry)
    nullable_args = (
        _extract_arg(name, entry_args[name]) for name in sorted(entry_args.keys())
    )
    args = filter(lambda item: item is not None, nullable_args)
    return list(args)


command_list: List[BuiltinEntry] = []



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.api.ctx** provides tools for preparing the project template.
"""

import datetime
import inspect
import os
from dataclasses import dataclass
from typing import Callable, Dict, Iterable, List, Optional, Union

from cxx_flow.base import cmd

package_root = os.path.dirname(os.path.dirname(__file__))
template_dir = "template"


StrOrBool = Union[str, bool]
Values = Union[StrOrBool, List[str]]
SettingsType = Dict[str, StrOrBool]


@dataclass
class Setting:
    json_key: str
    prompt: str = ""
    value: Union[Values, Callable[[], Values]] = ""
    fix: Optional[str] = None
    force_fix: bool = False

    def calc_value(self, previous: SettingsType):
        if isinstance(self.value, Callable):
            kwargs = {}

            params = inspect.signature(self.value).parameters
            if "settings" in params:
                kwargs["settings"] = previous

            return self.value(**kwargs)

        return self.value


def register_init_setting(*setting: Setting, is_hidden=False):
    (hidden if is_hidden else defaults).extend(setting)


def register_switch(key: str, prompt: str, enabled: bool):
    switches.append(Setting(key, prompt, value=enabled))


def register_internal(key: str, value: any):
    internals[key] = value


def _git_config(name: str):
    def wrap():
        proc = cmd.run("git", "config", name, capture_output=True)
        if proc is None or proc.returncode != 0:
            return None
        return proc.stdout.strip()

    return wrap


def move_to_front(preferred: str, values: Iterable[str]):
    result: List[str] = []

    has_preferred = False
    for value in values:
        if value == preferred:
            has_preferred = True
        else:
            result.append(value)

    if has_preferred:
        result.insert(0, preferred)

    return result


def _list_ext():
    return move_to_front(".cpp", sorted(_fileext.keys()))


def _as_mustache(basename: str):
    name, ext = os.path.splitext(basename)
    return ext == ".mustache" and name or None


def _enum_licenses():
    root = os.path.abspath(os.path.join(package_root, template_dir, "licenses"))
    for _, dirnames, filenames in os.walk(root):
        dirnames[:] = []
        iter = filter(lambda x: x is not None, map(_as_mustache, filenames))
        return move_to_front("MIT", iter)
    return []


def _filter(
    src: Callable[[SettingsType], StrOrBool],
    flt: Callable[[StrOrBool], StrOrBool],
):
    def impl(settings: SettingsType):
        return flt(src(settings))

    return impl


def _get_nothing(_: SettingsType) -> StrOrBool:
    return ""


def _map(internal_key: str):
    def impl(key: str):
        mapped = internals.get(internal_key)
        if not isinstance(mapped, dict):
            return None
        return mapped.get(key)

    return impl


def _get_key(key: str):
    def impl(settings: SettingsType):
        return settings.get(key, "")

    return impl


def _build_fixup(settings: SettingsType, fixup: str):
    fixup_refs = fixup.split("{")
    result = fixup_refs.pop(0)
    for ref_expr in fixup_refs:
        ref_text, verbose = ref_expr.split("}")
        filter_path = ref_text.split("$")

        ref = filter_path.pop(0)
        code = _get_nothing if ref == "" else _get_key(ref)

        for filter_name in filter_path:
            flt = _filters.get(filter_name)
            if flt is None:
                if filter_name.startswith("map:"):
                    internal_key = filter_name[4:]
                    code = _filter(code, _map(internal_key))
                continue
            code = _filter(code, flt)

        value = code(settings)

        if result:
            result += value
        else:
            result = value

        if verbose:
            result += verbose

    return result


def _fixed(fixup: str):
    def wrap(settings: SettingsType):
        return _build_fixup(settings, fixup)

    return wrap


internals = {}

switches: List[Setting] = []

defaults: List[Setting] = [
    Setting(
        "PROJECT.NAME",
        "Project name",
        lambda: os.path.basename(os.path.abspath("")) or "?project-name?",
    ),
    Setting("PROJECT.DESCRIPTION", "Project description"),
    Setting(
        "PROJECT.EMAIL",
        "Valid email, e.g. for CODE_OF_CONDUCT",
        _git_config("user.email"),
    ),
    Setting(
        "COPY.YEAR",
        "Year for copyright notices",
        str(datetime.date.today().year),
    ),
    Setting(
        "COPY.HOLDER",
        "Holder of the copyright",
        _git_config("user.name"),
    ),
    Setting("COPY.LICENSE", "License", _enum_licenses),
    Setting(
        "INCLUDE_PREFIX",
        'Prefix for includes (as in #include "{PREFIX}/version.hpp")',
        _fixed("{PROJECT.NAME}"),
    ),
    Setting(
        "NAME_PREFIX",
        "CMake variable name prefix",
        _fixed("{PROJECT.NAME$safe$upper}"),
    ),
    Setting(
        "NAMESPACE",
        "C++ namespace for the project",
        _fixed("{PROJECT.NAME$safe}"),
    ),
    Setting("EXT", "Extension for code files", _list_ext),
    Setting("SRCDIR", "Directory for code files", "src"),
    Setting(
        "INCLUDEDIR",
        "Directory for include files",
        "include",
        "{INCLUDEDIR}/{INCLUDE_PREFIX}",
        force_fix=True,
    ),
]

hidden: List[Setting] = [
    Setting("EXT.cxx", fix="{EXT}"),
    Setting("EXT.hxx", fix="{EXT.cxx$header}"),
]

_fileext = {".cc": ".hh", ".cxx": ".hxx", ".cpp": ".hpp"}


_filters: Dict[str, Callable[[StrOrBool], StrOrBool]] = {
    "safe": lambda value: value.replace("-", "_"),
    "upper": lambda value: value.upper(),
    "lower": lambda value: value.lower(),
    "header": lambda cxx_ext: _fileext.get(cxx_ext, ".hpp"),
}



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.api.step** exposes APIs used by run extensions.
"""

from abc import ABC, abstractmethod
from typing import List, cast

from cxx_flow.api.env import Config, Runtime
from cxx_flow.base import matrix


class Step(ABC):
    @property
    @abstractmethod
    def name(self) -> str: ...

    @property
    def runs_after(self) -> List[str]:
        return []

    @property
    def runs_before(self) -> List[str]:
        return []

    def platform_dependencies(self) -> List[str]:
        return []

    def is_active(self, config: Config, rt: Runtime) -> bool:
        return True

    def directories_to_remove(self, config: Config) -> List[str]:
        return []

    @abstractmethod
    def run(self, config: Config, rt: Runtime) -> int: ...


class SerialStep(Step):
    children: List[Step] = []

    @property
    def runs_after(self):
        return matrix.flatten([child.runs_after for child in self.children])

    def platform_dependencies(self) -> List[str]:
        return matrix.flatten(
            [child.platform_dependencies() for child in self.children]
        )

    def is_active(self, config: Config, rt: Runtime) -> bool:
        for child in self.children:
            if not child.is_active(config, rt):
                return False
        return True

    def directories_to_remove(self, config: Config) -> List[str]:
        return matrix.flatten(
            [child.directories_to_remove(config) for child in self.children]
        )

    def run(self, config: Config, rt: Runtime) -> int:
        for child in self.children:
            result = child.run(config, rt)
            if result:
                return result
        return 0


__steps: List[Step] = []


def _register_step(step: Step):
    global __steps

    name = step.name
    if name in [step.name for step in __steps]:
        raise NameError(f"Step {name} already registered")

    __steps.append(step)


def _inherits_from(base, existing_class):
    bases = list(existing_class.__bases__)
    for b in bases:
        if b == base or _inherits_from(base, b):
            return True
    return False


def _extend(existing_class, base):
    if _inherits_from(base, existing_class):
        return existing_class

    bases = list(existing_class.__bases__)
    bases.insert(0, base)

    new_class_namespace = existing_class.__dict__.copy()
    del new_class_namespace["__dict__"]

    metaclass = getattr(existing_class, "__metaclass__", type)
    return metaclass(existing_class.__name__, tuple(bases), new_class_namespace)


def _name_list(label: str, names: List[str], template="`{}`") -> str:
    if len(names) == 0:
        return ""

    em = [template.format(name) for name in names]
    prefix = ", ".join(em[:-1])
    if prefix:
        prefix += " and "
    return f"\n:{label}: {prefix}{em[-1]}"


def _make_private(f: callable):
    if f.__doc__:
        f.__doc__ += "\n\n:meta private:\n"
    else:
        f.__doc__ = ":meta private:\n"


_dummy_config = Config(
    {
        "os": "${os}",
        "build_type": "${build_type}",
        "build_name": "${build_name}",
        "preset": "${preset}",
        "build_generator": "${build_generator}",
    },
    [],
)


def _extend_docstring(conv, step: Step):
    info = "".join(
        (
            f"\n:Name: `{step.name}`",
            _name_list("Runs after", step.runs_after),
            _name_list("Runs before", step.runs_before),
            _name_list("Requires", step.platform_dependencies(), template="``{}``"),
            _name_list(
                "Removes",
                step.directories_to_remove(_dummy_config),
                template="``{}``",
            ),
        )
    )

    doc = conv.__doc__ or "*Docstring is missing!*"
    conv.__doc__ = f"{doc}\n{info}"

    _make_private(conv.is_active)
    _make_private(conv.run)
    _make_private(conv.platform_dependencies)
    _make_private(conv.directories_to_remove)


def register(cls=None):
    def impl(cls):
        conv = _extend(cls, Step)
        step = cast(Step, conv())
        _register_step(step)
        _extend_docstring(conv, step)

        return conv

    if cls is None:
        return impl

    return impl(cls)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.flow.cli.finder** extends the :py:class:`argcomplete.finders.CompletionFinder`
to be able to see proper :class:`api.env.FlowConfig`, when completed invocation
contains ``-C`` argument.
"""

import os
import shlex
from typing import List

from argcomplete import finders


def split_line(cli_text: str):
    last_is_finished = len(cli_text) > len(cli_text.rstrip())
    split = shlex.shlex(cli_text, posix=True)
    split.whitespace_split = True
    args = list(split)
    if not last_is_finished and len(args) > 2 and args[-2] == "-C":
        args = args[:-2]

    return args


def find_dash_see(args: List[str]):
    pos = -1
    max_pos = len(args)
    if max_pos:
        max_pos -= 1
    for index in range(max_pos):
        if args[index] == "-C":
            pos = index + 1

    return pos


class CompletionFinder(finders.CompletionFinder):
    def active(self):
        return "_ARGCOMPLETE" in os.environ

    def find_project(self):
        if not self.active():
            return "."

        comp_line = os.environ["COMP_LINE"]
        comp_point = int(os.environ["COMP_POINT"])
        args = split_line(comp_line[:comp_point])
        pos = find_dash_see(args)

        if pos < 0:
            return "."

        return args[pos]


autocomplete = CompletionFinder()



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.sign.win32** provides code signing with SignTool
from Windows SDKs.
"""

import base64
import json
import os
import platform
import struct
import subprocess
import sys
import winreg
from typing import Iterable, List, NamedTuple, Optional, Tuple

from cxx_flow.api.env import Runtime

ENV_KEY = "SIGN_TOKEN"

Version = Tuple[int, int, int]

machine = {"ARM64": "arm64", "AMD64": "x64", "X86": "x86"}.get(
    platform.machine(), "x86"
)


def find_sign_tool(rt: Runtime) -> Optional[str]:
    with winreg.OpenKeyEx(
        winreg.HKEY_LOCAL_MACHINE, r"SOFTWARE\Microsoft\Windows Kits\Installed Roots"
    ) as kits:
        try:
            kits_root = winreg.QueryValueEx(kits, "KitsRoot10")[0]
        except FileNotFoundError:
            rt.message("sign/win32: No KitsRoot10 value")
            return None

        versions: List[Tuple[Version, str]] = []
        try:
            index = 0
            while True:
                ver_str = winreg.EnumKey(kits, index)
                ver = tuple(int(chunk) for chunk in ver_str.split("."))
                index += 1
                versions.append((ver, ver_str))
        except OSError:
            pass
    versions.sort()
    versions.reverse()
    rt.message(
        "sign/win32: Regarding versions:", ", ".join(version[1] for version in versions)
    )
    for _, version in versions:
        # C:\Program Files (x86)\Windows Kits\10\bin\10.0.22621.0\x64\signtool.exe
        sign_tool = os.path.join(kits_root, "bin", version, machine, "signtool.exe")
        if os.path.isfile(sign_tool):
            rt.message("sign/win32: using:", sign_tool)
            return sign_tool
    return None


class Key(NamedTuple):
    token: str
    secret: bytes


def _get_key_from_contents(key: str, rt: Runtime):
    try:
        obj = json.loads(key)
    except json.decoder.JSONDecodeError:
        rt.message("sign/win32: the signature is not a valid JSON document")
        return None

    if not isinstance(obj, dict):
        rt.message("sign/win32: the signature is missing required fields")
        return None

    token = obj.get("token")
    secret = obj.get("secret")
    if not isinstance(token, str) or not isinstance(secret, str):
        rt.message("sign/win32: the signature is missing required fields")
        return None

    return Key(
        base64.b64decode(token).decode("UTF-8"),
        base64.b64decode(secret),
    )


def get_key(rt: Runtime) -> Optional[Key]:
    rt.message(f"sign/win32: trying ${ENV_KEY}")
    env = os.environ.get(ENV_KEY)
    if env:
        key = _get_key_from_contents(env, rt)
        if key is not None:
            return key
    local_signature = os.path.join(".", "signature.key")
    home_signature = os.path.join(os.path.expanduser("~"), "signature.key")
    for filename in [local_signature, home_signature]:
        rt.message(f"sign/win32: trying {filename}")
        if os.path.isfile(filename):
            with open(filename, encoding="UTF-8") as file:
                result = file.read().strip()
                key = _get_key_from_contents(result, rt)
                if key is not None:
                    return key

    rt.message("sign/win32: no key set up")

    return None


def is_active(os_name: str, rt: Runtime):
    if os_name != "windows":
        return False
    key = get_key(rt)
    return (
        key is not None
        and key.token is not None
        and key.secret is not None
        and find_sign_tool(rt) is not None
    )


_IMAGE_DOS_HEADER = "HHHHHHHHHHHHHH8sHH20sI"
_IMAGE_NT_HEADERS_Signature = "H"
_IMAGE_DOS_HEADER_size = struct.calcsize(_IMAGE_DOS_HEADER)
_IMAGE_NT_HEADERS_Signature_size = struct.calcsize(_IMAGE_NT_HEADERS_Signature)
_MZ = 23117
_PE = 17744


def is_pe_exec(path: str):
    with open(path, "rb") as exe:
        mz_header = exe.read(_IMAGE_DOS_HEADER_size)
        dos_header = struct.unpack(_IMAGE_DOS_HEADER, mz_header)
        if dos_header[0] != _MZ:
            return False

        PE_offset = dos_header[-1]
        if PE_offset < _IMAGE_DOS_HEADER_size:
            return False

        if PE_offset > _IMAGE_DOS_HEADER_size:
            exe.read(PE_offset - _IMAGE_DOS_HEADER_size)

        pe_header = exe.read(_IMAGE_NT_HEADERS_Signature_size)
        signature = struct.unpack(_IMAGE_NT_HEADERS_Signature, pe_header)[0]
        return signature == _PE


def sign(files: Iterable[str], rt: Runtime):
    key = get_key(rt)

    if key is None or key.token is None or key.secret is None:
        print("cxx-flow: sign: the key is missing", file=sys.stderr)
        return 1

    sign_tool = find_sign_tool(rt)
    if sign_tool is None:
        print("cxx-flow: sign: signtool.exe not found", file=sys.stderr)
        sys.exit(0)

    with open("temp.pfx", "wb") as pfx:
        pfx.write(key.secret)

    args = [
        sign_tool,
        "sign",
        "/f",
        "temp.pfx",
        "/p",
        key.token,
        "/tr",
        "http://timestamp.digicert.com",
        "/fd",
        "sha256",
        "/td",
        "sha256",
        *files,
    ]

    result = 1
    try:
        result = subprocess.run(args, shell=False).returncode
    finally:
        os.remove("temp.pfx")

    return result



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.commands.ci** implements ``./flow ci`` command.
"""

from cxx_flow.api import arg, env

from . import matrix

__all__ = ["matrix", "command_ci"]


@arg.command("ci")
def main():
    """Perform various CI tasks"""



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.api** contains public APIs, usable in third-party plugins.
"""

from . import arg, completers, ctx, env, init, makefile, step

__all__ = ["arg", "completers", "ctx", "env", "init", "makefile", "step"]



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.base.matrix** operates on lists of
:class:`run configurations <cxx_flow.api.env.Config>`. It provides means to
build cartesian products for multiple sets, to load the GitHub Actions Job
strategy matrix from a set of YAML files, or to put additional filters on the
resulting set of configurations.
"""

import os
import sys
from typing import Dict, List, Tuple, TypeVar

import yaml

T = TypeVar("T")


def find_compiler(
    compiler: str, config_names: Dict[str, List[str]]
) -> Tuple[str, List[str]]:
    """
    Locates the C and C++ names of compilers corresponding to the current
    compiler.

    :param compiler: Name ofthe compiuler to map, taken from the ``"compiler"``
        entry in run :class:`config<cxx_flow.api.env.Config>`.
    :param config_names: Dictionary of mapping from compiler name to C/C++
        names, taken from ``"compiler.names"`` entry in flow config file.
    """
    dirname = os.path.dirname(compiler)
    filename = os.path.basename(compiler)
    if sys.platform == "win32":
        filename = os.path.splitext(filename)[0]
    chunks = filename.split("-", 1)
    if len(chunks) == 1:
        version = None
    else:
        version = chunks[1]
    filename = chunks[0].lower()

    try:
        compiler_names = config_names[filename]
    except:
        compiler_names = [filename]

    compilers = [
        os.path.join(dirname, name if version is None else f"{name}-{version}")
        for name in compiler_names
    ]

    if filename == "stdclang":
        filename = "clang"

    return filename, compilers


def flatten(array: List[List[T]]) -> List[T]:
    """
    Turns list of lists into a list.

    :param array: List of lists to flatten
    :returns: a list with all items expanded
    """
    return [item for sublist in array for item in sublist]


def matches(tested: dict, test: dict) -> bool:
    """
    Checks, if the tested dictionary contains all the values from test
    dictionary.

    :param tested: Dictionary to check
    :param test: Dictionary to check against

    :returns: `True`, if all keys from `test` are in `tested` and have the same
        values, `False` otherwise.
    """

    for key, value in test.items():
        val = tested.get(key)
        if val != value:
            return False
    return True


def matches_any(tested: dict, tests: List[dict]):
    """
    Checks, if the tested dictionary contains all the values from at least one
    of test dictionaries.

    :param tested: Dictionary to check
    :param tests: List of dictionaries to check against

    :returns: `True`, if at least one test :func:`matches` `tested`
        dictionary, `False` otherwise.
    """

    for test in tests:
        if matches(tested, test):
            return True
    return False


def cartesian(input: Dict[str, list]) -> List[dict]:
    """
    Calculates the cartesian product of all axes in `input`.

    The input dictionary is source of all axes this product is calculated from.
    The key to the dictionary names the set and value under each key represents
    the set to be multiplied.

    The output list contains each possible dictionary, where key set is
    identical to the input key set and each value is one of the values in the
    input axis.

    :param input: A dictionary with each axis and its values
    :returns: A list of all possible permutation of axes values; order in which
        each permutation appears in resulting list is not guaranteed.

    Example:
    --------

    .. code-block:: python
        :caption: Input

        {
            "key-1": ["value-1", "value-2"],
            "key-2": [True, False],
            "key-3": [1, 2, 3],
        }

    .. code-block:: python
        :caption: Possible output

        [
            { "key-1": "value-1", "key-2": [True], "key-3": 1 },
            { "key-1": "value-1", "key-2": [True], "key-3": 2 },
            { "key-1": "value-1", "key-2": [True], "key-3": 3 },
            { "key-1": "value-1", "key-2": [False], "key-3":  1 },
            { "key-1": "value-1", "key-2": [False], "key-3":  2 },
            { "key-1": "value-1", "key-2": [False], "key-3":  3 },
            { "key-1": "value-2", "key-2": [True], "key-3":  1 },
            { "key-1": "value-2", "key-2": [True], "key-3":  2 },
            { "key-1": "value-2", "key-2": [True], "key-3":  3 },
            { "key-1": "value-2", "key-2": [False], "key-3":  1 },
            { "key-1": "value-2", "key-2": [False], "key-3":  2 },
            { "key-1": "value-2", "key-2": [False], "key-3":  3 },
        ]
    """
    product = [{}]

    for key, values in input.items():
        next_level = []
        for value in values:
            for obj in product:
                next_level.append({**obj, key: value})
        product = next_level

    return product


def _split_keys(includes: List[dict], keys: List[str]) -> List[Tuple[dict, dict]]:
    result = []
    for include in includes:
        expand_key = {}
        expand_value = {}
        for key, value in include.items():
            if key in keys:
                expand_key[key] = value
            else:
                expand_value[key] = value
        result.append((expand_key, expand_value))
    return result


def load_matrix(*matrix_paths: str) -> Tuple[List[dict], List[str]]:
    """
    Loads config definitions from files under `matrix_paths` to produce
    cartesian product of the matrix, expanded and/or contracted according to
    rules used by GitHub Action workflow `strategies`_. Please note when
    comparing to GitHub documentation:

    - only scalar values are supported in the ``matrix`` dictionary.
    - only ``matrix``, ``include`` and ``exclude`` keys are allowed.

    :param matrix_paths: The list of filenames pointing to YAML files with
        ``matrix``, ``include`` and ``exclude`` entries.
    :returns: A tuple of cartesian product for the config matrix, with
        inclusions and exclusions as described in GitHub docs and a list of key
        names in the ``matrix`` dictionary.

    .. _strategies: https://docs.github.com/en/actions/writing-workflows/workflow-syntax-for-github-actions#jobsjob_idstrategy
    """
    setups: List[dict] = []
    for matrix_path in matrix_paths:
        try:
            with open(matrix_path, encoding="UTF-8") as f:
                setups.append(yaml.load(f, Loader=yaml.Loader))
        except FileNotFoundError:
            pass

    if len(setups) == 0:
        return [], []

    setup = setups[0]
    for additional in setups[1:]:
        src_matrix = setup.get("matrix", {})
        src_exclude = setup.get("exclude", [])
        src_include = setup.get("include", [])

        for key, value in additional.get("matrix", {}).items():
            old = src_matrix.get(key)
            if isinstance(old, list) and isinstance(value, list):
                old.extend(value)
            elif isinstance(old, list):
                old.append(value)
            else:
                src_matrix[key] = value
        src_exclude.extend(additional.get("exclude", []))
        src_include.extend(additional.get("include", []))

    raw = setup.get("matrix", {})
    keys = list(raw.keys())
    full = cartesian(raw)

    includes = _split_keys(setup.get("include", []), keys)
    for obj in full:
        for include_key, include_value in includes:
            if not matches(obj, include_key):
                continue
            for key, value in include_value.items():
                obj[key] = value

    excludes = setup.get("exclude", [])
    matrix = [obj for obj in full if not matches_any(obj, excludes)]

    return matrix, keys



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.conan** provides the ``"Conan"`` step.
"""

import os
import textwrap
from typing import List

from cxx_flow.api import ctx, env, step

from ._conan import conan_api

CONAN_DIR = "build/conan"
CONAN_PROFILE = "_profile-compiler"
CONAN_PROFILE_GEN = "_profile-build_type"


@step.register
class ConanConfig:
    """Configures the project for ``preset`` config using ``build_type`` config."""

    name = "Conan"

    def platform_dependencies(self):
        return ["conan"]

    def is_active(self, config: env.Config, rt: env.Runtime) -> int:
        return os.path.isfile("conanfile.txt") or os.path.isfile("conanfile.py")

    def directories_to_remove(self, _: env.Config) -> List[str]:
        return [CONAN_DIR]

    def run(self, config: env.Config, rt: env.Runtime) -> int:
        api = conan_api()

        profile_gen = f"{CONAN_DIR}/{CONAN_PROFILE_GEN}-{config.preset}"
        if not rt.dry_run:
            os.makedirs(CONAN_DIR, exist_ok=True)
            with open(profile_gen, "w", encoding="UTF-8") as profile:
                print(
                    textwrap.dedent(
                        f"""\
                        include({CONAN_PROFILE})

                        [settings]"""
                    ),
                    file=profile,
                )

                for setting in [
                    *api.settings(config),
                    f"build_type={config.build_type}",
                ]:
                    print(setting, file=profile)

        if api.config(rt, CONAN_DIR, f"./{CONAN_DIR}/{CONAN_PROFILE}", profile_gen):
            return 1
        if not rt.dry_run and os.path.exists("CMakeUserPresets.json"):
            os.remove("CMakeUserPresets.json")
        return 0


ctx.register_switch("with_conan", "Use Conan for dependency manager", True)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.cmake.config** provides ``"CMake"`` step.
"""

import os
from typing import Dict, List, cast

from cxx_flow import api

from .__version__ import CMAKE_VERSION


@api.step.register
class CMakeConfig:
    """Configures the project using ``preset`` config."""

    name = "CMake"
    runs_after = ["Conan"]

    def platform_dependencies(self):
        return [f"cmake>={CMAKE_VERSION}"]

    def is_active(self, config: api.env.Config, rt: api.env.Runtime) -> int:
        return os.path.isfile("CMakeLists.txt") and os.path.isfile("CMakePresets.json")

    def directories_to_remove(self, config: api.env.Config) -> List[str]:
        return [f"build/{config.build_type}"]

    def run(self, config: api.env.Config, rt: api.env.Runtime) -> int:
        cmake_vars = cast(Dict[str, str], rt._cfg.get("cmake", {}).get("vars", {}))
        defines: List[str] = []
        for var in cmake_vars:
            value = cmake_vars[var]

            is_flag = value.startswith("?")
            if is_flag:
                value = value[1:]

            if value.startswith("config:"):
                value = value[len("config:")]
                value = config.get_path(value)
            elif value.startswith("runtime:"):
                value = value[len("runtime:")]
                value = getattr(rt, value, None)

            if is_flag:
                value = "ON" if value else "OFF"

            defines.append(f"-D{var}={value}")

        return rt.cmd(
            "cmake",
            "--preset",
            f"{config.preset}-{config.build_generator}",
            *defines,
        )



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.flow.interact** provides initialization context through user
prompts.
"""

from dataclasses import dataclass
from typing import List, Union

from prompt_toolkit import prompt as tk_prompt
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.shortcuts import CompleteStyle
from prompt_toolkit.validation import Validator

from cxx_flow.api import ctx


@dataclass
class _Question:
    key: str
    prompt: str
    value: ctx.Values

    @classmethod
    def load_default(cls, default: ctx.Setting, previous: ctx.SettingsType):
        value = default.calc_value(previous)
        return cls(default.json_key, default.prompt, value)

    def interact(self, counter: int, size: int) -> ctx.StrOrBool:
        if isinstance(self.value, str):
            return self._get_str(self.value, counter, size)
        if isinstance(self.value, bool):
            return self._get_flag(self.value, counter, size)
        return self._get_list(self.value, counter, size)

    @property
    def ps(self):
        return self.prompt or f'"{self.key}"'

    def _ps(self, default: ctx.Values, counter: int, size: int):
        if default:
            if isinstance(default, str):
                return [
                    ("", f"[{counter}/{size}] {self.ps} ["),
                    ("bold", default),
                    ("", f"]: "),
                ]
            if isinstance(default, bool):
                b = "bold"
                n = ""
                on_true = (b if default else n, "yes")
                on_false = (b if not default else n, "no")
                return [
                    ("", f"[{counter}/{size}] {self.ps} ["),
                    on_true,
                    ("", " / "),
                    on_false,
                    ("", f"]: "),
                ]
            return [
                ("", f"[{counter}/{size}] {self.ps} ["),
                ("bold", default[0]),
                ("", f"{''.join(f' / {x}' for x in default[1:])}]: "),
            ]
        return f"[{counter}/{size}] {self.ps}: "

    def _get_str(self, default: str, counter: int, size: int):
        value = tk_prompt(self._ps(default, counter, size))
        if not value:
            value = default
        return value

    def _get_flag(self, default: bool, counter: int, size: int):
        value = self._tk_prompt(
            default, ["yes", "no", "on", "off", "1", "0"], counter, size
        )
        result = default
        if value:
            result = value.lower() in ["1", "on", "y", "yes"]
        return result

    def _get_list(self, defaults: List[str], counter: int, size: int):
        value = self._tk_prompt(defaults, defaults, counter, size)
        if not value:
            value = defaults[0]
        return value

    def _tk_prompt(
        self,
        defaults: Union[bool | List[str]],
        words: List[str],
        counter: int,
        size: int,
    ):
        def valid(word: str):
            return word == "" or word in words

        validator = Validator.from_callable(valid)
        completer = WordCompleter(words)
        return tk_prompt(
            self._ps(defaults, counter, size),
            validator=validator,
            completer=completer,
            complete_while_typing=True,
            complete_style=CompleteStyle.READLINE_LIKE,
        )


def prompt() -> ctx.SettingsType:
    """
    Prompts user to provide details of newly-crated project.

    :returns: Dictionary with answers to all interactive settings and switches.
    """
    settings: ctx.SettingsType = {}

    size = len(ctx.defaults) + len(ctx.switches)
    counter = 1

    for setting in ctx.defaults:
        loaded = _Question.load_default(setting, settings)
        value = loaded.interact(counter, size)
        settings[loaded.key] = value
        counter += 1

    for setting in ctx.switches:
        loaded = _Question.load_default(setting, settings)
        value = loaded.interact(counter, size)
        settings[loaded.key] = value
        counter += 1

    return settings



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.base** contains low-level tools for higher-level parts of the
library.
"""

from . import cmd, matrix, plugins, uname

__all__ = ["cmd", "matrix", "plugins", "uname"]



# PYTHON_ARGCOMPLETE_OK

# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.flow.cli** provides command-line entry for the *C++ flow*.
"""

import argparse
import os
import sys
from typing import List, cast

from cxx_flow import __version__
from cxx_flow.api import env
from cxx_flow.flow import steps
from cxx_flow.flow.cli import cmds, finder


def _change_dir():
    root = argparse.ArgumentParser(
        prog="cxx-flow",
        usage="cxx-flow [-h] [--version] [-C [dir]] {command} ...",
        add_help=False,
    )
    root.add_argument("-C", dest="cd", nargs="?")

    args, _ = root.parse_known_args()
    if args.cd:
        os.chdir(args.cd)


def _expand_shortcuts(parser: argparse.ArgumentParser, args: argparse.Namespace):
    args_kwargs = dict(args._get_kwargs())
    for key in parser.shortcuts:
        try:
            if not args_kwargs[key]:
                continue
            cast(List[List[str]], args.configs).append(parser.shortcuts[key])
            break
        except KeyError:
            continue


def __main():
    _change_dir()

    flow_cfg = env.FlowConfig(root=finder.autocomplete.find_project())
    steps.clean_aliases(flow_cfg)

    parser = cmds.build_argparser(flow_cfg)
    finder.autocomplete(parser)
    args = parser.parse_args()
    _expand_shortcuts(parser, args)

    sys.exit(cmds.BuiltinEntry.run_entry(args, flow_cfg))


def main():
    """Entry point for *C++ flow* tool."""
    try:
        __main()
    except KeyboardInterrupt:
        sys.exit(1)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.cmake.pack** provides ``"Pack"`` step.
"""

import os

from cxx_flow.api import env, step

from .__version__ import CMAKE_VERSION


@step.register
class PackStep:
    """
    Packs archives and installers from ``cpack_generator`` config, using
    ``preset`` config.
    """

    name = "Pack"
    runs_after = ["Build"]

    def platform_dependencies(self):
        return [f"cmake>={CMAKE_VERSION}", f"cpack>={CMAKE_VERSION}"]

    def is_active(self, config: env.Config, rt: env.Runtime) -> int:
        return (
            os.path.isfile("CMakeLists.txt")
            and os.path.isfile("CMakePresets.json")
            and len(config.items.get("cpack_generator", [])) > 0
        )

    def run(self, config: env.Config, rt: env.Runtime) -> int:
        generators = ";".join(config.items.get("cpack_generator", []))
        return rt.cmd("cpack", "--preset", config.preset, "-G", generators)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.__main__** allows *C++ flow* to be called as Python component.
"""

from .flow.cli import main

main()



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins** provides the runtime comnponents of the *C++ flow*:
the :class:`initialization helpers <cxx_flow.api.init.InitStep>` and
:class:`run steps <cxx_flow.api.step.Step>`.
"""



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.store** provides ``"StoreTests"`` step.
"""

from cxx_flow.api import env, step


@step.register
class StoreTests(step.Step):
    """Stores test results gathered during tests for ``preset`` config value."""

    name = "StoreTests"
    runs_after = ["Test"]

    def run(self, config: env.Config, rt: env.Runtime) -> int:
        return rt.cp(
            f"build/{config.preset}/test-results", "build/artifacts/test-results"
        )



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.sign** provides the ``"Sign"`` and ``"SignPackages"``
steps.
"""

import fnmatch
import os
import sys
from abc import abstractmethod
from typing import List, cast

from cxx_flow.api import env, init, step

if sys.platform == "win32":
    from . import win32

else:

    class win32:
        @staticmethod
        def is_active(*args):
            return False

        @staticmethod
        def sign(*args):
            return 0

        @staticmethod
        def is_pe_exec(arg):
            return False


def should_exclude(filename: str, exclude: List[str], config_os: str):
    basename = os.path.splitext(filename)[0] if config_os == "windows" else filename

    for pattern in exclude:
        if fnmatch.fnmatch(basename, pattern):
            return True

    return False


class SignBase(step.Step):
    def is_active(self, config: env.Config, rt: env.Runtime) -> int:
        return win32.is_active(config.os, rt)

    @abstractmethod
    def get_files(self, config: env.Config, rt: env.Runtime) -> List[str]: ...

    def run(self, config: env.Config, rt: env.Runtime) -> int:
        files = [file.replace(os.sep, "/") for file in self.get_files(config, rt)]
        if len(files) == 0:
            return 0

        rt.print("signtool", *(os.path.basename(file) for file in files))

        if rt.dry_run:
            return 0

        return win32.sign(files, rt)


@step.register
class SignFiles(SignBase):
    """*(Windows)* Signs executable files in build directory"""

    name = "Sign"
    runs_after = ["Build"]
    runs_before = ["Pack"]

    def get_files(self, config: env.Config, rt: env.Runtime) -> List[str]:
        cfg = cast(dict, rt._cfg.get("sign", {}))
        roots = cfg.get("directories", ["bin", "lib", "libexec", "share"])
        exclude = cfg.get("exclude", ["*-test"])

        result: List[str] = []
        build_dir = config.build_dir
        for root in roots:
            for curr_dir, _, filenames in os.walk(os.path.join(build_dir, root)):
                for filename in filenames:
                    if should_exclude(filename, exclude, config.os):
                        continue

                    full_path = os.path.join(curr_dir, filename)
                    if not win32.is_pe_exec(full_path):
                        continue

                    result.append(full_path)
        return result


@step.register
class SignMsi(SignBase):
    """*(Windows)* Signs MSI installers in build directory"""

    name = "SignPackages"
    runs_after = ["Pack"]
    runs_before = ["StorePackages", "Store"]

    def is_active(self, config: env.Config, rt: env.Runtime) -> int:
        return super().is_active(config, rt) and "WIX" in config.items.get(
            "cpack_generator", []
        )

    def get_files(self, config: env.Config, rt: env.Runtime) -> List[str]:
        result: List[str] = []
        pkg_dir = os.path.join(config.build_dir, "packages")
        for curr_dir, dirnames, filenames in os.walk(pkg_dir):
            dirnames[:] = []
            for filename in filenames:
                _, ext = os.path.splitext(filename)
                if ext.lower() != ".msi":
                    continue

                result.append(os.path.join(curr_dir, filename))

        return result


class SignInit(init.InitStep):
    def postprocess(self, rt: env.Runtime, context: dict):
        if sys.platform == "win32":
            with open(".gitignore", "ab") as ignoref:
                ignoref.write("\n/signature.key\n".encode("UTF-8"))


init.register_init_step(SignInit())



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.base.uname** provides platform information usable in packaging.
"""

import platform
from typing import Optional, Tuple

NameStr = str
VersionStr = Optional[str]
ArchStr = str


def uname() -> Tuple[NameStr, VersionStr, ArchStr]:
    """
    Return normalized information from :py:mod:`platform` module.

    Specifically, the machine architecture is swapped from "amd64" to "x86_64",
    which is more common in package filenames.

    On Windows system, with OS not easily returning current version and with
    high level of backward compatibility, system version is removed
    all-together.

    On Linux systems, true name of the system is retrieved, using
    :py:func:`platform.freedesktop_os_release`. On systems, where Python is
    missing this function, a fallback is made to "linux" and version returned
    by :py:func:`platform.uname`.

    :returns: a tuple consisting of system name, version and
       machine architecture

    :see: :ref:`command-system` command
    """
    _platform = platform.uname()
    platform_name = _platform.system.lower()
    platform_version = _platform.version
    platform_arch = _platform.machine.lower()

    if platform_arch == "amd64":
        platform_arch = "x86_64"

    system_nt = platform_name.split("_nt-", 1)

    if len(system_nt) > 1:
        platform_name = system_nt[0]
        platform_version = None
    elif platform_name == "windows":
        platform_version = None
    elif platform_name == "linux":
        try:
            os_release = platform.freedesktop_os_release()
        except AttributeError:
            os_release = {"ID": "linux", "NAME": "Linux"}

        platform_id = os_release.get("ID", os_release.get("NAME"))
        version_id = os_release.get("VERSION_ID", platform_version)
        if platform_id is not None:
            if platform_id[:9] == "opensuse-":
                platform_id = "opensuse"
            platform_name = platform_id.lower()
            if version_id != "" and version_id[0] in "0123456789":
                platform_version = ".".join(version_id.split(".", 2)[:2])

    return (
        platform_name,
        platform_version,
        platform_arch,
    )



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.flow.dependency** verifies availabilty of Step's external tools.
"""

import re
import sys
from dataclasses import dataclass
from enum import Enum
from importlib.metadata import version as package_version
from typing import Callable, List, Set, Tuple

from cxx_flow.base import cmd

VER_REGEX = re.compile(r"((?:0|[1-9][0-9]*)\.(?:0|[1-9][0-9]*)\.(?:0|[1-9][0-9]*))")
Version = Tuple[int, int, int]


class DepKind(Enum):
    APP = "app"
    PYTHON_PKG = "python_pkg"


def _ver(ver: str) -> Version:
    chunks = [int(v.strip()) for v in ver.split(".")]
    while len(chunks) < 3:
        chunks.append(0)
    return (*chunks[:3],)


@dataclass
class Comp:
    op: str
    check: Callable[[Version, Version], bool]


def _compatible(app: Version, exp: Version):
    return app[0] == exp[0] and app[1] >= exp[1]


_comparisons: List[Comp] = [
    Comp("==", lambda app, exp: app == exp),
    Comp(">=", lambda app, exp: app >= exp),
    Comp("<=", lambda app, exp: app <= exp),
    Comp("~=", _compatible),
    Comp("!=", lambda app, exp: app != exp),
    Comp("<", lambda app, exp: app < exp),
    Comp(">", lambda app, exp: app > exp),
]


@dataclass
class Dependency:
    name: str
    version_expression: str
    kind: DepKind = DepKind.APP

    def __lt__(self, rhs: "Dependency"):
        if self.kind != rhs.kind:
            return self.kind.value < rhs.kind.value
        if self.name != rhs.name:
            return self.name < rhs.name
        return self.version_expression < rhs.version_expression

    def __eq__(self, rhs: "Dependency"):
        return (
            self.kind == rhs.kind
            and self.name == rhs.name
            and self.version_expression == rhs.version_expression
        )

    def match_version(self, version: str):
        for expression in (expr.strip() for expr in self.version_expression.split(",")):
            if not self._match_version(version, expression):
                return f"{self.name}: version `{version}` does not match `{expression}`"
        return None

    def _match_version(self, version: str, expression: str):
        if expression == "":
            return True
        app = _ver(version)
        for comp in _comparisons:
            if not expression.startswith(comp.op):
                continue
            return comp.check(app, _ver(expression[len(comp.op) :].lstrip()))

        return False  # there is some other permutation of "=><!~"


def gather(steps: list):
    deps: Set[str] = set()
    for step in steps:
        deps.update(step.platform_dependencies())

    result: List[Dependency] = []
    for dep in deps:
        split = re.split(r"([=><!~])", dep, 1)
        app = split.pop(0).strip()
        version = "".join(split).strip()

        kind = DepKind.APP
        app_keys = app.split()
        if len(app_keys) > 2 and app_keys[0] == "python" and app_keys[1] == "-m":
            app = " ".join(app_keys[2:])
            kind = DepKind.PYTHON_PKG

        result.append(Dependency(app, version, kind))

    return result


def verify(deps: List[Dependency]):
    uniq: List[Dependency] = []
    errors: Set[str] = set()
    for dep in sorted(deps):
        if len(uniq) == 0 or uniq[-1] != dep:
            uniq.append(dep)

    for pkg in (dep for dep in uniq if dep.kind == DepKind.PYTHON_PKG):
        try:
            version = package_version(pkg.name)
        except:
            errors.add(f"{pkg.name}: Python package is missing")
            continue
        msg = pkg.match_version(version)
        if msg is not None:
            errors.add(msg)

    for app in (dep for dep in uniq if dep.kind == DepKind.APP):
        if not cmd.is_tool(app.name):
            errors.add(f"{app.name}: tool is missing")
            continue
        proc = cmd.run(app.name, "--version", capture_output=True)
        if proc.returncode:
            if proc.stderr:
                print(proc.stderr.rstrip(), file=sys.stderr)
            version = None
        else:
            m = VER_REGEX.search(proc.stdout)
            version = m.group(0) if m is not None else None

        if version is None and app.version_expression != "":
            errors.add(
                f"{app.name}: could not read version for `{app.version_expression}`"
            )
            continue
        msg = app.match_version(version)
        if msg is not None:
            errors.add(msg)

    return list(sorted(errors))



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.commands.ci.matrix** implements ``./flow ci matrix`` command.
"""

import argparse
import json
import os
import sys
from typing import Annotated, Optional

from cxx_flow.api import arg, env
from cxx_flow.flow.configs import Configs


@arg.command("ci", "matrix")
def matrix(
    official: Annotated[
        bool, arg.FlagArgument(help="Cut matrix to release builds only")
    ],
    rt: env.Runtime,
):
    """Supply data for github actions"""

    configs = Configs(
        rt, argparse.Namespace(configs=[], matrix=True, official=official)
    )

    usable = [usable.items for usable in configs.usable]
    for config in usable:
        if "--orig-compiler" in config:
            orig_compiler = config["--orig-compiler"]
            del config["--orig-compiler"]
            config["compiler"] = orig_compiler
    if "GITHUB_ACTIONS" in os.environ:
        var = json.dumps({"include": usable})
        GITHUB_OUTPUT = os.environ.get("GITHUB_OUTPUT")
        if GITHUB_OUTPUT is not None:
            with open(GITHUB_OUTPUT, "a", encoding="UTF-8") as github_output:
                print(f"matrix={var}", file=github_output)
        else:
            print(f"matrix={var}")
    else:
        json.dump(usable, sys.stdout)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.flow.steps** allows loading both predefined and project-specific
steps.
"""

import importlib
import os
import sys
from dataclasses import dataclass
from types import ModuleType
from typing import List, Optional, Union, cast

from cxx_flow.api import env, step
from cxx_flow.base.plugins import load_module_plugins


@dataclass
class Sorted:
    plugin: step.Step
    name: str
    runs_after: List[str]
    runs_before: List[str]

    @staticmethod
    def from_step(plugin: step.Step):
        name = plugin.name
        runs_after = [*plugin.runs_after]
        runs_before = [*plugin.runs_before]
        return Sorted(plugin, name, runs_after, runs_before)


def _sort_steps():
    steps = step.__steps
    unsorted = [Sorted.from_step(step) for step in steps]
    known_names = [step.name for step in unsorted]

    for plugin in unsorted:
        for name in plugin.runs_before:
            for successor in unsorted:
                if successor.name != name:
                    continue
                successor.runs_after.append(plugin.name)

    for plugin in unsorted:
        runs_after: List[str] = []
        for name in plugin.runs_after:
            if name in known_names:
                runs_after.append(name)
        plugin.runs_after = runs_after

    result: List[step.Step] = []

    while len(unsorted) > 0:
        layer = [plugin for plugin in unsorted if len(plugin.runs_after) == 0]
        unsorted = [plugin for plugin in unsorted if len(plugin.runs_after) > 0]
        for plugin in layer:
            for remaining in unsorted:
                try:
                    remaining.runs_after.remove(plugin.name)
                except ValueError:
                    pass
        result.extend(plugin.plugin for plugin in layer)
        if len(layer) == 0:
            result.extend(plugin.plugin for plugin in unsorted)
            break

    return result


def clean_aliases(cfg: env.FlowConfig):
    entries = cfg.entry
    if not entries:
        return

    valid_steps = _sort_steps()
    step_names = {step.name for step in valid_steps}

    keys_to_remove: List[str] = []
    for key in entries:
        entry = entries[key]
        if isinstance(entry, dict):
            steps = cast(List[str], entry.get("steps", []))
        else:
            steps = cast(List[str], entry)
        if len(steps) == 0:
            keys_to_remove.append(key)
            continue
        rewritten: List[str] = []
        for step in steps:
            if step in step_names:
                rewritten.append(step)
        steps[:] = rewritten

        if not steps:
            keys_to_remove.append(key)

    for key in keys_to_remove:
        del entries[key]

    cfg.steps = valid_steps
    cfg.aliases = [env.RunAlias.from_json(key, entries[key]) for key in entries]



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.commands** package implements various CLI commands.
"""

import sys

from cxx_flow.base.plugins import load_module_plugins

load_module_plugins(sys.modules[__name__])



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.plugins.store** provides ``"Store"``, ``"StoreTests"`` and
``"StorePackages"`` steps.
"""

from . import store_both, store_packages, store_tests

__all__ = ["store_both", "store_tests", "store_packages"]



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **cxx_flow.base.plugins** provide the plugin enumeration helpers.
"""

import importlib
import os
from types import ModuleType
from typing import Optional


def _load_plugins(directory: str, package: Optional[str], can_fail=False):
    for _, dirnames, filenames in os.walk(directory):
        for dirname in dirnames:
            if dirname == "__pycache__":
                continue

            try:
                importlib.import_module(f".{dirname}", package=package)
            except ModuleNotFoundError as err:
                if not can_fail:
                    raise err
        for filename in filenames:
            if filename == "__init__.py":
                continue

            try:
                importlib.import_module(
                    f".{os.path.splitext(filename)[0]}", package=package
                )
            except ModuleNotFoundError as err:
                if not can_fail:
                    raise err
        dirnames[:] = []


def load_module_plugins(mod: ModuleType, can_fail=False):
    spec = mod.__spec__
    if not spec:
        return
    for location in spec.submodule_search_locations:
        _load_plugins(location, spec.name, can_fail)
